# snapshottest: v1 - https://goo.gl/zC4yUc

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots[
    "test_atomic_diagnostic_observed_value_expect_column_kl_divergence_to_be_less_than 1"
] = {
    "name": "atomic.diagnostic.observed_value",
    "value": {
        "graph": {
            "$schema": "https://vega.github.io/schema/vega-lite/v4.8.1.json",
            "autosize": "fit",
            "config": {"view": {"continuousHeight": 300, "continuousWidth": 400}},
            "data": {"name": "data-c49f2d3d7bdab36f9ec81f5a314a430c"},
            "datasets": {
                "data-c49f2d3d7bdab36f9ec81f5a314a430c": [
                    {"fraction": 0.3754, "values": 1},
                    {"fraction": 0.615, "values": 2},
                    {"fraction": 0.0096, "values": 4},
                ]
            },
            "encoding": {
                "tooltip": [
                    {"field": "values", "type": "quantitative"},
                    {"field": "fraction", "type": "quantitative"},
                ],
                "x": {"field": "values", "type": "nominal"},
                "y": {"field": "fraction", "type": "quantitative"},
            },
            "height": 400,
            "mark": "bar",
            "width": 250,
        },
        "header": {
            "schema": {"type": "StringValueType"},
            "value": {
                "params": {
                    "observed_value": {
                        "schema": {"type": "string"},
                        "value": "None (-infinity, infinity, or NaN)",
                    }
                },
                "template": "KL Divergence: $observed_value",
            },
        },
        "schema": {"type": "GraphType"},
    },
    "value_type": "GraphType",
}

snapshots[
    "test_atomic_diagnostic_observed_value_with_boolean_columun_expect_column_kl_divergence_to_be_less_than 1"
] = {
    "name": "atomic.diagnostic.observed_value",
    "value": {
        "graph": {
            "$schema": "https://vega.github.io/schema/vega-lite/v4.8.1.json",
            "autosize": "fit",
            "config": {"view": {"continuousHeight": 300, "continuousWidth": 400}},
            "data": {"name": "data-d8f1a1ab1f79e142d9ca399157673554"},
            "datasets": {
                "data-d8f1a1ab1f79e142d9ca399157673554": [
                    {"fraction": 0.5, "values": "True"},
                    {"fraction": 0.5, "values": "False"},
                ]
            },
            "encoding": {
                "tooltip": [
                    {"field": "values", "type": "nominal"},
                    {"field": "fraction", "type": "quantitative"},
                ],
                "x": {"field": "values", "type": "nominal"},
                "y": {"field": "fraction", "type": "quantitative"},
            },
            "height": 400,
            "mark": "bar",
            "width": 250,
        },
        "header": {
            "schema": {"type": "StringValueType"},
            "value": {
                "params": {
                    "observed_value": {
                        "schema": {"type": "string"},
                        "value": "None (-infinity, infinity, or NaN)",
                    }
                },
                "template": "KL Divergence: $observed_value",
            },
        },
        "schema": {"type": "GraphType"},
    },
    "value_type": "GraphType",
}

snapshots[
    "test_atomic_diagnostic_observed_value_expect_column_quantile_values_to_be_between 1"
] = {
    "name": "atomic.diagnostic.observed_value",
    "value": {
        "header_row": [
            {"schema": {"type": "string"}, "value": "Quantile"},
            {"schema": {"type": "string"}, "value": "Value"},
        ],
        "schema": {"type": "TableType"},
        "table": [
            [
                {"schema": {"type": "string"}, "value": "0.05"},
                {"schema": {"type": "number"}, "value": 67},
            ],
            [
                {"schema": {"type": "string"}, "value": "Q1"},
                {"schema": {"type": "number"}, "value": 329},
            ],
            [
                {"schema": {"type": "string"}, "value": "Median"},
                {"schema": {"type": "number"}, "value": 657},
            ],
            [
                {"schema": {"type": "string"}, "value": "Q3"},
                {"schema": {"type": "number"}, "value": 985},
            ],
            [
                {"schema": {"type": "string"}, "value": "0.95"},
                {"schema": {"type": "number"}, "value": 1247},
            ],
        ],
    },
    "value_type": "TableType",
}

snapshots["test_atomic_diagnostic_observed_value_with_empty_result 1"] = {
    "name": "atomic.diagnostic.observed_value",
    "value": {
        "params": {},
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "--",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_diagnostic_observed_value_with_numeric_observed_value 1"] = {
    "name": "atomic.diagnostic.observed_value",
    "value": {
        "params": {},
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "1,776",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_diagnostic_observed_value_with_str_observed_value 1"] = {
    "name": "atomic.diagnostic.observed_value",
    "value": {
        "params": {},
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "foo",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_diagnostic_observed_value_with_unexpected_percent 1"] = {
    "name": "atomic.diagnostic.observed_value",
    "value": {
        "params": {},
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "10% unexpected",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_diagnostic_observed_value_without_result 1"] = {
    "name": "atomic.diagnostic.observed_value",
    "value": {
        "params": {},
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "--",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_distinct_values_to_be_in_set 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "v__0": {"schema": {"type": "number"}, "value": 1},
            "v__1": {"schema": {"type": "number"}, "value": 2},
            "v__2": {"schema": {"type": "number"}, "value": 3},
            "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3]},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column distinct values must belong to this set: $v__0 $v__1 $v__2.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_distinct_values_to_contain_set 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "parse_strings_as_datetimes": {
                "schema": {"type": "boolean"},
                "value": True,
            },
            "v__0": {"schema": {"type": "string"}, "value": "a"},
            "v__1": {"schema": {"type": "string"}, "value": "b"},
            "v__2": {"schema": {"type": "string"}, "value": "c"},
            "value_set": {"schema": {"type": "array"}, "value": ["a", "b", "c"]},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column distinct values must contain this set: $v__0 $v__1 $v__2. Values should be parsed as datetimes.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_distinct_values_to_equal_set 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "parse_strings_as_datetimes": {
                "schema": {"type": "boolean"},
                "value": True,
            },
            "v__0": {"schema": {"type": "string"}, "value": "a"},
            "v__1": {"schema": {"type": "string"}, "value": "b"},
            "v__2": {"schema": {"type": "string"}, "value": "c"},
            "value_set": {"schema": {"type": "array"}, "value": ["a", "b", "c"]},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column distinct values must match this set: $v__0 $v__1 $v__2. Values should be parsed as datetimes.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_kl_divergence_to_be_less_than 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "graph": {
            "$schema": "https://vega.github.io/schema/vega-lite/v4.8.1.json",
            "autosize": "fit",
            "config": {"view": {"continuousHeight": 300, "continuousWidth": 400}},
            "data": {"name": "data-1cb20570b53cc3e67cb4883fd45e64cb"},
            "datasets": {
                "data-1cb20570b53cc3e67cb4883fd45e64cb": [
                    {"bin_max": 5, "bin_min": 0, "fraction": 0.2},
                    {"bin_max": 10, "bin_min": 5, "fraction": 0.3},
                    {"bin_max": 30, "bin_min": 10, "fraction": 0.1},
                    {"bin_max": 50, "bin_min": 30, "fraction": 0.4},
                ]
            },
            "encoding": {
                "tooltip": [
                    {"field": "bin_min", "type": "quantitative"},
                    {"field": "bin_max", "type": "quantitative"},
                    {"field": "fraction", "type": "quantitative"},
                ],
                "x": {"field": "bin_min", "type": "ordinal"},
                "x2": {"field": "bin_max"},
                "y": {"field": "fraction", "type": "quantitative"},
            },
            "height": 400,
            "mark": "bar",
            "width": 250,
        },
        "header": {
            "schema": {"type": "StringValueType"},
            "value": {
                "params": {
                    "column": {"schema": {"type": "string"}, "value": "min_event_time"},
                    "threshold": {"schema": {"type": "number"}, "value": 0.1},
                },
                "template": "$column Kullback-Leibler (KL) divergence with respect to the following distribution must be lower than $threshold.",
            },
        },
        "schema": {"type": "GraphType"},
    },
    "value_type": "GraphType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_max_to_be_between 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 5},
            "min_value": {"schema": {"type": "number"}, "value": 1},
            "parse_strings_as_datetimes": {
                "schema": {"type": "boolean"},
                "value": True,
            },
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column maximum value must be greater than or equal to $min_value and less than or equal to $max_value. Values should be parsed as datetimes.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_mean_to_be_between 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 7},
            "min_value": {"schema": {"type": "number"}, "value": 3},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column mean must be greater than or equal to $min_value and less than or equal to $max_value.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_median_to_be_between 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 10},
            "min_value": {"schema": {"type": "number"}, "value": 5},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column median must be greater than or equal to $min_value and less than or equal to $max_value.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_min_to_be_between 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 5},
            "min_value": {"schema": {"type": "number"}, "value": 1},
            "parse_strings_as_datetimes": {
                "schema": {"type": "boolean"},
                "value": True,
            },
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column minimum value must be greater than or equal to $min_value and less than or equal to $max_value. Values should be parsed as datetimes.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_most_common_value_to_be_in_set 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "ties_okay": {"schema": {"type": "boolean"}, "value": True},
            "v__0": {"schema": {"type": "number"}, "value": 1},
            "v__1": {"schema": {"type": "number"}, "value": 2},
            "v__2": {"schema": {"type": "number"}, "value": 3},
            "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3]},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column most common value must belong to this set: $v__0 $v__1 $v__2. Values outside this set that are as common (but not more common) are allowed.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_pair_cramers_phi_value_to_be_less_than 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column_A": {"schema": {"type": "string"}, "value": "foo"},
            "column_B": {"schema": {"type": "string"}, "value": "bar"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Values in $column_A and $column_B must be independent.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_pair_values_a_to_be_greater_than_b 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column_A": {"schema": {"type": "string"}, "value": "foo"},
            "column_B": {"schema": {"type": "string"}, "value": "bar"},
            "ignore_row_if": {"schema": {"type": "string"}, "value": "baz"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "or_equal": {"schema": {"type": "boolean"}, "value": True},
            "parse_strings_as_datetimes": {
                "schema": {"type": "boolean"},
                "value": True,
            },
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Values in $column_A must be greater than or equal to those in $column_B, at least $mostly_pct % of the time. Values should be parsed as datetimes.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_pair_values_to_be_equal 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column_A": {"schema": {"type": "string"}, "value": "foo"},
            "column_B": {"schema": {"type": "string"}, "value": "bar"},
            "ignore_row_if": {"schema": {"type": "string"}, "value": "baz"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Values in $column_A and $column_B must be equal, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_proportion_of_unique_values_to_be_between 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 20},
            "min_value": {"schema": {"type": "number"}, "value": 10},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column fraction of unique values must be greater than or equal to $min_value and less than or equal to $max_value.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_quantile_values_to_be_between 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "header": {
            "schema": {"type": "StringValueType"},
            "value": {
                "params": {
                    "column": {"schema": {"type": "string"}, "value": "Unnamed: 0"},
                },
                "template": "$column quantiles must be within the following value ranges.",
            },
        },
        "header_row": [
            {"schema": {"type": "string"}, "value": "Quantile"},
            {"schema": {"type": "string"}, "value": "Min Value"},
            {"schema": {"type": "string"}, "value": "Max Value"},
        ],
        "schema": {"type": "TableType"},
        "table": [
            [
                {"schema": {"type": "string"}, "value": "0.05"},
                {"schema": {"type": "number"}, "value": 66},
                {"schema": {"type": "number"}, "value": 68},
            ],
            [
                {"schema": {"type": "string"}, "value": "Q1"},
                {"schema": {"type": "number"}, "value": 328},
                {"schema": {"type": "number"}, "value": 330},
            ],
            [
                {"schema": {"type": "string"}, "value": "Median"},
                {"schema": {"type": "number"}, "value": 656},
                {"schema": {"type": "number"}, "value": 658},
            ],
            [
                {"schema": {"type": "string"}, "value": "Q3"},
                {"schema": {"type": "number"}, "value": 984},
                {"schema": {"type": "number"}, "value": 986},
            ],
            [
                {"schema": {"type": "string"}, "value": "0.95"},
                {"schema": {"type": "number"}, "value": 1246},
                {"schema": {"type": "number"}, "value": 1248},
            ],
        ],
    },
    "value_type": "TableType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_stdev_to_be_between 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 20},
            "min_value": {"schema": {"type": "number"}, "value": 10},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column standard deviation must be greater than or equal to $min_value and less than or equal to $max_value.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_sum_to_be_between 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 20},
            "min_value": {"schema": {"type": "number"}, "value": 10},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column sum must be greater than or equal to $min_value and less than or equal to $max_value.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_to_exist 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "column_index": {"schema": {"type": "number"}, "value": 5},
            "column_indexth": {"schema": {"type": "string"}, "value": "5th"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column must be the $column_indexth field.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_unique_value_count_to_be_between 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 20},
            "min_value": {"schema": {"type": "number"}, "value": 10},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column must have greater than or equal to $min_value and less than or equal to $max_value unique values, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_value_lengths_to_be_between 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 20},
            "min_value": {"schema": {"type": "number"}, "value": 10},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be greater than or equal to $min_value and less than or equal to $max_value characters long, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_value_lengths_to_equal 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "value": {"schema": {"type": "number"}, "value": 100},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be $value characters long, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_values_to_be_between 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "max_value": {"schema": {"type": "number"}, "value": 5},
            "min_value": {"schema": {"type": "number"}, "value": 1},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be greater than or equal to $min_value and less than or equal to $max_value, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_be_dateutil_parseable 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be parseable by dateutil, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_be_decreasing 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "parse_strings_as_datetimes": {
                "schema": {"type": "boolean"},
                "value": True,
            },
            "strictly": {"schema": {"type": "boolean"}, "value": True},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be strictly less than previous values, at least $mostly_pct % of the time. Values should be parsed as datetimes.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_values_to_be_in_set 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "parse_strings_as_datetimes": {
                "schema": {"type": "boolean"},
                "value": True,
            },
            "v__0": {"schema": {"type": "number"}, "value": 1},
            "v__1": {"schema": {"type": "number"}, "value": 2},
            "v__2": {"schema": {"type": "number"}, "value": 3},
            "v__3": {"schema": {"type": "number"}, "value": 4},
            "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3, 4]},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must belong to this set: $v__0 $v__1 $v__2 $v__3, at least $mostly_pct % of the time. Values should be parsed as datetimes.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_be_in_type_list 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "type_list": {
                "schema": {"type": "array"},
                "value": ["type_a", "type_b", "type_c"],
            },
            "v__0": {"schema": {"type": "string"}, "value": "type_a"},
            "v__1": {"schema": {"type": "string"}, "value": "type_b"},
            "v__2": {"schema": {"type": "string"}, "value": "type_c"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column value types must belong to this set: $v__0 $v__1 $v__2, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_be_increasing 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "parse_strings_as_datetimes": {
                "schema": {"type": "boolean"},
                "value": True,
            },
            "strictly": {"schema": {"type": "boolean"}, "value": True},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be strictly greater than previous values, at least $mostly_pct % of the time. Values should be parsed as datetimes.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_be_json_parseable 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be parseable as JSON, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_values_to_be_null 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be null, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_be_null_with_mostly_equals_1 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 1.0},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be null.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_values_to_be_of_type 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "type_": {"schema": {"type": "string"}, "value": "my_type"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be of type $type_, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_values_to_be_unique 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must be unique, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_match_json_schema 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "json_schema": {"schema": {"type": "object"}, "value": {"foo": "bar"}},
            "formatted_json": {
                "schema": {"type": "string"},
                "value": "<pre>{\n" '    "foo": "bar"\n' "}</pre>",
            },
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must match the following JSON Schema, at least $mostly_pct % of the time: $formatted_json",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_values_to_match_regex 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "regex": {"schema": {"type": "string"}, "value": "^superconductive$"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must match this regular expression: $regex, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_match_regex_list 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "match_on": {"schema": {"type": "string"}, "value": "all"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "regex_list": {
                "schema": {"type": "array"},
                "value": ["^superconductive$", "ge|great_expectations"],
            },
            "v__0": {"schema": {"type": "string"}, "value": "^superconductive$"},
            "v__1": {"schema": {"type": "string"}, "value": "ge|great_expectations"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must match all of the following regular expressions: $v__0 $v__1, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_match_strftime_format 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "strftime_format": {"schema": {"type": "string"}, "value": "%Y-%m"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must match the following strftime format: $strftime_format, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_not_be_in_set 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "v__0": {"schema": {"type": "number"}, "value": 1},
            "v__1": {"schema": {"type": "number"}, "value": 2},
            "v__2": {"schema": {"type": "number"}, "value": 3},
            "value_set": {"schema": {"type": "array"}, "value": [1, 2, 3]},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must not belong to this set: $v__0 $v__1 $v__2, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_column_values_to_not_be_null 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must not be null, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_not_match_regex 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "regex": {"schema": {"type": "string"}, "value": "^superconductive$"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must not match this regular expression: $regex, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_column_values_to_not_match_regex_list 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column": {"schema": {"type": "string"}, "value": "my_column"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "regex_list": {"schema": {"type": "array"}, "value": ["^a", "^b", "^c"]},
            "v__0": {"schema": {"type": "string"}, "value": "^a"},
            "v__1": {"schema": {"type": "string"}, "value": "^b"},
            "v__2": {"schema": {"type": "string"}, "value": "^c"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "$column values must not match any of the following regular expressions: $v__0 $v__1 $v__2, at least $mostly_pct % of the time.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_compound_columns_to_be_unique 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column_list": {
                "schema": {"type": "array"},
                "value": ["my_first_col", "my_second_col", "my_third_col"],
            },
            "column_list_0": {"schema": {"type": "string"}, "value": "my_first_col"},
            "column_list_1": {"schema": {"type": "string"}, "value": "my_second_col"},
            "column_list_2": {"schema": {"type": "string"}, "value": "my_third_col"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Values for given compound columns must be unique together, at least $mostly_pct % of the time: $column_list_0 $column_list_1 $column_list_2",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_multicolumn_values_to_be_unique 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column_list": {"schema": {"type": "array"}, "value": ["A", "B", "C"]},
            "ignore_row_if": {"schema": {"type": "string"}, "value": "foo"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "column_list_0": {"schema": {"type": "string"}, "value": "A"},
            "column_list_1": {"schema": {"type": "string"}, "value": "B"},
            "column_list_2": {"schema": {"type": "string"}, "value": "C"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Values must be unique across columns, at least $mostly_pct % of the time: $column_list_0 $column_list_1 $column_list_2",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_select_column_values_to_be_unique_within_record 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column_list": {
                "schema": {"type": "array"},
                "value": ["my_first_column", "my_second_column"],
            },
            "ignore_row_if": {"schema": {"type": "string"}, "value": "foo"},
            "mostly": {"schema": {"type": "number"}, "value": 0.8},
            "mostly_pct": {"schema": {"type": "string"}, "value": "80"},
            "column_list_0": {"schema": {"type": "string"}, "value": "my_first_column"},
            "column_list_1": {
                "schema": {"type": "string"},
                "value": "my_second_column",
            },
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Values must be unique across columns, at least $mostly_pct % of the time: $column_list_0 $column_list_1",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_table_column_count_to_be_between 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "min_value": {"schema": {"type": "number"}, "value": 5},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Must have greater than or equal to $min_value columns.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_table_column_count_to_equal 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {"value": {"schema": {"type": "number"}, "value": 10}},
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Must have exactly $value columns.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_table_columns_to_match_ordered_list 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column_list": {"schema": {"type": "array"}, "value": ["a", "b", "c"]},
            "column_list_0": {"schema": {"type": "string"}, "value": "a"},
            "column_list_1": {"schema": {"type": "string"}, "value": "b"},
            "column_list_2": {"schema": {"type": "string"}, "value": "c"},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Must have these columns in this order: $column_list_0 $column_list_1 $column_list_2",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_table_columns_to_match_set 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "column_set": {"schema": {"type": "array"}, "value": ["a", "b", "c"]},
            "column_set_0": {"schema": {"type": "string"}, "value": "a"},
            "column_set_1": {"schema": {"type": "string"}, "value": "b"},
            "column_set_2": {"schema": {"type": "string"}, "value": "c"},
            "exact_match": {"schema": {"type": "boolean"}, "value": True},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Must have exactly these columns (in any order): $column_set_0 $column_set_1 $column_set_2",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_table_row_count_to_be_between 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "min_value": {"schema": {"type": "number"}, "value": 1},
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Must have greater than or equal to $min_value rows.",
    },
    "value_type": "StringValueType",
}

snapshots["test_atomic_prescriptive_summary_expect_table_row_count_to_equal 1"] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {"value": {"schema": {"type": "number"}, "value": 10}},
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Must have exactly $value rows.",
    },
    "value_type": "StringValueType",
}

snapshots[
    "test_atomic_prescriptive_summary_expect_table_row_count_to_equal_other_table 1"
] = {
    "name": "atomic.prescriptive.summary",
    "value": {
        "params": {
            "other_table_name": {
                "schema": {"type": "string"},
                "value": {"schema": {"type": "string"}, "value": "other_table_name"},
            }
        },
        "schema": {"type": "com.superconductive.rendered.string"},
        "template": "Row count must equal the row count of table $other_table_name.",
    },
    "value_type": "StringValueType",
}



from typing import List

from ruamel import yaml

import great_expectations as gx
from great_expectations.core.batch import Batch, BatchRequest

context = gx.get_context()

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/cloud/gcs/pandas/configured_yaml_example.py datasource config">
datasource_yaml = f"""
name: my_gcs_datasource
class_name: Datasource
execution_engine:
    class_name: PandasExecutionEngine
data_connectors:
    configured_data_connector_name:
        class_name: ConfiguredAssetGCSDataConnector
        bucket_or_name: <YOUR_GCS_BUCKET_HERE>
        prefix: <BUCKET_PATH_TO_DATA>
        default_regex:
            pattern: data/taxi_yellow_tripdata_samples/yellow_tripdata_sample_(\\d{{4}})-(\\d{{2}})\\.csv
            group_names:
                - year
                - month
        assets:
            taxi_data:
"""
# </snippet>

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your path directly in the yaml above.
datasource_yaml = datasource_yaml.replace("<YOUR_GCS_BUCKET_HERE>", "test_docs_data")
datasource_yaml = datasource_yaml.replace(
    "<BUCKET_PATH_TO_DATA>", "data/taxi_yellow_tripdata_samples/"
)

context.test_yaml_config(datasource_yaml)

# <snippet name="tests/integration/docusaurus/connecting_to_your_data/cloud/gcs/pandas/configured_yaml_example.py test datasource config">
context.add_datasource(**yaml.load(datasource_yaml))
# </snippet>

# Here is a BatchRequest naming a data_asset
batch_request = BatchRequest(
    datasource_name="my_gcs_datasource",
    data_connector_name="configured_data_connector_name",
    data_asset_name="<YOUR_DATA_ASSET_NAME>",
)

# Please note this override is only to provide good UX for docs and tests.
# In normal usage you'd set your data asset name directly in the BatchRequest above.
batch_request.data_asset_name = "taxi_data"

context.add_or_update_expectation_suite(expectation_suite_name="test_suite")
validator = context.get_validator(
    batch_request=batch_request, expectation_suite_name="test_suite"
)
print(validator.head())


# NOTE: The following code is only for testing and can be ignored by users.
assert isinstance(validator, gx.validator.validator.Validator)
assert [ds["name"] for ds in context.list_datasources()] == ["my_gcs_datasource"]
assert set(
    context.get_available_data_asset_names()["my_gcs_datasource"][
        "configured_data_connector_name"
    ]
) == {"taxi_data"}

batch_list: List[Batch] = context.get_batch_list(batch_request=batch_request)
assert len(batch_list) == 3

batch: Batch = batch_list[0]
assert batch.data.dataframe.shape[0] == 10000



from typing import Optional, cast

import pandas as pd
import pytest

from great_expectations import DataContext
from great_expectations.core.expectation_validation_result import (
    ExpectationValidationResult,
)
from great_expectations.self_check.util import (
    build_pandas_validator_with_data,
    build_sa_validator_with_data,
    get_test_validator_with_data,
)
from great_expectations.util import build_in_memory_runtime_context, is_library_loadable


@pytest.mark.skipif(
    not is_library_loadable(library_name="pyathena"),
    reason="pyathena is not installed",
)
def test_expect_column_values_to_be_in_type_list_dialect_pyathena_string(sa):
    from pyathena import sqlalchemy_athena

    df = pd.DataFrame({"col": ["test_val1", "test_val2"]})
    validator = build_sa_validator_with_data(df, "sqlite")

    # Monkey-patch dialect for testing purposes.
    validator.execution_engine.dialect_module = sqlalchemy_athena

    result = validator.expect_column_values_to_be_in_type_list(
        "col", type_list=["string", "boolean"]
    )

    assert result == ExpectationValidationResult(
        success=True,
        expectation_config={
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "col",
                "type_list": ["string", "boolean"],
            },
            "meta": {},
        },
        result={
            "element_count": 2,
            "unexpected_count": 0,
            "unexpected_percent": 0.0,
            "partial_unexpected_list": [],
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent_total": 0.0,
            "unexpected_percent_nonmissing": 0.0,
        },
        exception_info={
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
        meta={},
    )


@pytest.mark.skipif(
    not is_library_loadable(library_name="pyathena"),
    reason="pyathena is not installed",
)
def test_expect_column_values_to_be_in_type_list_dialect_pyathena_boolean(sa):
    from pyathena import sqlalchemy_athena

    df = pd.DataFrame({"col": [True, False]})
    validator = build_sa_validator_with_data(df, "sqlite")

    # Monkey-patch dialect for testing purposes.
    validator.execution_engine.dialect_module = sqlalchemy_athena

    result = validator.expect_column_values_to_be_in_type_list(
        "col", type_list=["string", "boolean"]
    )

    assert result == ExpectationValidationResult(
        success=True,
        expectation_config={
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "col",
                "type_list": ["string", "boolean"],
            },
            "meta": {},
        },
        result={
            "element_count": 2,
            "unexpected_count": 0,
            "unexpected_percent": 0.0,
            "partial_unexpected_list": [],
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent_total": 0.0,
            "unexpected_percent_nonmissing": 0.0,
        },
        exception_info={
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
        meta={},
    )


def test_expect_column_values_to_be_in_type_list_nullable_int():
    from packaging.version import parse

    pandas_version = parse(pd.__version__)
    if pandas_version < parse("0.24"):
        # Prior to 0.24, Pandas did not have
        pytest.skip("Prior to 0.24, Pandas did not have `Int32Dtype` or related.")

    df = pd.DataFrame({"col": pd.Series([1, 2, None], dtype=pd.Int32Dtype())})

    context: Optional[DataContext] = cast(
        DataContext, build_in_memory_runtime_context()
    )
    validator = get_test_validator_with_data(
        execution_engine="pandas",
        data=df,
        context=context,
    )

    result = validator.expect_column_values_to_be_in_type_list(
        "col", type_list=["Int32Dtype"]
    )
    assert result == ExpectationValidationResult(
        success=True,
        expectation_config={
            "expectation_type": "expect_column_values_to_be_in_type_list",
            "kwargs": {
                "column": "col",
                "type_list": ["Int32Dtype"],
            },
            "meta": {},
        },
        result={
            "element_count": 3,
            "unexpected_count": 0,
            "unexpected_percent": 0.0,
            "partial_unexpected_list": [],
            "missing_count": 0,
            "missing_percent": 0.0,
            "unexpected_percent_total": 0.0,
            "unexpected_percent_nonmissing": 0.0,
        },
        exception_info={
            "raised_exception": False,
            "exception_traceback": None,
            "exception_message": None,
        },
        meta={},
    )



from numbers import Number
from typing import Dict, Optional

import numpy as np
import pandas as pd
import pytest
import scipy.stats as stats

RANDOM_SEED: int = 43792
RANDOM_STATE: np.random.Generator = np.random.Generator(np.random.PCG64(RANDOM_SEED))


def generate_distribution_parameters(
    false_positive_rate: np.float64,
) -> Dict[str, Dict[str, Number]]:
    normal_mean: int = 5000
    normal_stdev: int = 1000
    normal_lower_quantile: Number = stats.norm.ppf(
        q=false_positive_rate / 2, loc=normal_mean, scale=normal_stdev
    )
    normal_upper_quantile: Number = stats.norm.ppf(
        q=1 - false_positive_rate / 2, loc=normal_mean, scale=normal_stdev
    )

    uniform_lower_bound: int = 4000
    uniform_scale: int = 6000
    uniform_lower_quantile: Number = stats.uniform.ppf(
        q=false_positive_rate / 2, loc=uniform_lower_bound, scale=uniform_scale
    )
    uniform_upper_quantile: Number = stats.uniform.ppf(
        q=1 - false_positive_rate / 2, loc=uniform_lower_bound, scale=uniform_scale
    )

    bimodal_mean_1: int = 4000
    bimodal_stdev_1: int = 500
    bimodal_mean_2: int = 6000
    bimodal_stdev_2: int = 500

    # 10,000 samples taken from a bimodal mixture of equally weighted normal distributions.
    # This approximation is being used due to the lack of functionality in scipy.stats regarding a
    # percent point functipn (ppf) for gaussian mixtures.
    # For a look at a similar statistical representation and implementation in R see:
    # https://stats.stackexchange.com/questions/390931/compute-quantile-function-from-a-mixture-of-normal-distribution
    bimodal_approximation = np.concatenate(
        [
            stats.norm.rvs(
                loc=bimodal_mean_1,
                scale=bimodal_stdev_1,
                size=5000,
                random_state=RANDOM_STATE,
            ),
            stats.norm.rvs(
                loc=bimodal_mean_2,
                scale=bimodal_stdev_2,
                size=5000,
                random_state=RANDOM_STATE,
            ),
        ]
    )
    bimodal_lower_quantile = np.quantile(
        a=bimodal_approximation, q=false_positive_rate / 2
    )
    bimodal_upper_quantile = np.quantile(
        a=bimodal_approximation, q=1 - false_positive_rate / 2
    )

    exponential_shape: float = 1.5
    exponential_lower_bound: int = 5000
    exponential_scale: int = 1000
    exponential_lower_quantile: Number = stats.gamma.ppf(
        q=false_positive_rate / 2,
        a=exponential_shape,
        loc=exponential_lower_bound,
        scale=exponential_scale,
    )
    exponential_upper_quantile: Number = stats.gamma.ppf(
        q=1 - false_positive_rate / 2,
        a=exponential_shape,
        loc=exponential_lower_bound,
        scale=exponential_scale,
    )

    return {
        "normal": {
            "mean": normal_mean,
            "stdev": normal_stdev,
            "lower_quantile": normal_lower_quantile,
            "upper_quantile": normal_upper_quantile,
        },
        "uniform": {
            "lower_bound": uniform_lower_bound,
            "scale": uniform_scale,
            "lower_quantile": uniform_lower_quantile,
            "upper_quantile": uniform_upper_quantile,
        },
        "bimodal": {
            "mean_1": bimodal_mean_1,
            "stdev_1": bimodal_stdev_1,
            "mean_2": bimodal_mean_2,
            "stdev_2": bimodal_stdev_2,
            "lower_quantile": bimodal_lower_quantile,
            "upper_quantile": bimodal_upper_quantile,
        },
        "exponential": {
            "shape": exponential_shape,
            "lower_bound": exponential_lower_bound,
            "scale": exponential_scale,
            "lower_quantile": exponential_lower_quantile,
            "upper_quantile": exponential_upper_quantile,
        },
    }


def generate_distribution_samples(
    distribution_parameters: Dict[str, Dict[str, Number]], size: Optional[int] = 36
) -> pd.DataFrame:
    data: Dict[str, np.ndarray] = {
        "normal": np.around(
            stats.norm.rvs(
                loc=distribution_parameters["normal"]["mean"],
                scale=distribution_parameters["normal"]["stdev"],
                size=size,
                random_state=RANDOM_STATE,
            )
        ),
        "uniform": np.around(
            stats.uniform.rvs(
                loc=distribution_parameters["uniform"]["lower_bound"],
                scale=distribution_parameters["uniform"]["scale"],
                size=size,
                random_state=RANDOM_STATE,
            )
        ),
        "bimodal": np.around(
            np.concatenate(
                [
                    stats.norm.rvs(
                        loc=distribution_parameters["bimodal"]["mean_1"],
                        scale=distribution_parameters["bimodal"]["stdev_1"],
                        size=size // 2,
                        random_state=RANDOM_STATE,
                    ),
                    stats.norm.rvs(
                        loc=distribution_parameters["bimodal"]["mean_2"],
                        scale=distribution_parameters["bimodal"]["stdev_2"],
                        size=size // 2,
                        random_state=RANDOM_STATE,
                    ),
                ]
            )
        ),
        "exponential": np.around(
            stats.gamma.rvs(
                a=distribution_parameters["exponential"]["shape"],
                loc=distribution_parameters["exponential"]["lower_bound"],
                scale=distribution_parameters["exponential"]["scale"],
                size=size,
                random_state=RANDOM_STATE,
            )
        ),
    }
    return pd.DataFrame(data)


@pytest.fixture
def bootstrap_distribution_parameters_and_1000_samples_with_01_false_positive():
    false_positive_rate: np.float64 = np.float64(0.01)
    distribution_parameters: Dict[
        str, Dict[str, Number]
    ] = generate_distribution_parameters(false_positive_rate=false_positive_rate)
    distribution_samples: pd.DataFrame = generate_distribution_samples(
        distribution_parameters=distribution_parameters, size=1000
    )
    return {
        "false_positive_rate": false_positive_rate,
        "distribution_parameters": distribution_parameters,
        "distribution_samples": distribution_samples,
    }



import os
import sys
from typing import List

import click

from great_expectations import DataContext  # noqa: TCH001
from great_expectations.checkpoint.types.checkpoint_result import (
    CheckpointResult,  # noqa: TCH001
)
from great_expectations.cli import toolkit
from great_expectations.cli.pretty_printing import cli_message, cli_message_list
from great_expectations.core.usage_statistics.events import UsageStatsEvents
from great_expectations.core.usage_statistics.util import send_usage_message
from great_expectations.data_context.util import file_relative_path
from great_expectations.exceptions import InvalidTopLevelConfigKeyError
from great_expectations.render.renderer.checkpoint_new_notebook_renderer import (
    CheckpointNewNotebookRenderer,
)
from great_expectations.util import lint_code

try:
    from sqlalchemy.exc import SQLAlchemyError
except ImportError:
    SQLAlchemyError = RuntimeError


try:
    from sqlalchemy.exc import SQLAlchemyError
except ImportError:
    SQLAlchemyError = RuntimeError


"""
--ge-feature-maturity-info--

    id: checkpoint_command_line
    title: LegacyCheckpoint - Command Line
    icon:
    short_description: Run a configured Checkpoint from a command line.
    description: Run a configured Checkpoint from a command line in a Terminal shell.
    how_to_guide_url: https://docs.greatexpectations.io/en/latest/guides/how_to_guides/validation/how_to_run_a_checkpoint_in_terminal.html
    maturity: Experimental
    maturity_details:
        api_stability: Unstable (expect changes to batch request)
        implementation_completeness: Complete
        unit_test_coverage: Complete
        integration_infrastructure_test_coverage: N/A
        documentation_completeness: Complete
        bug_risk: Low

--ge-feature-maturity-info--
"""


@click.group(short_help="Checkpoint operations")
@click.pass_context
def checkpoint(ctx: click.Context) -> None:
    """
    Checkpoint operations

    A Checkpoint is a bundle of one or more batches of data with one or more
    Expectation Suites.

    A Checkpoint can be as simple as one batch of data paired with one
    Expectation Suite.

    A Checkpoint can be as complex as many batches of data across different
    datasources paired with one or more Expectation Suites each.
    """
    ctx.obj.data_context = ctx.obj.get_data_context_from_config_file()

    cli_event_noun: str = "checkpoint"
    (
        begin_event_name,
        end_event_name,
    ) = UsageStatsEvents.get_cli_begin_and_end_event_names(
        noun=cli_event_noun,
        verb=ctx.invoked_subcommand,
    )
    send_usage_message(
        data_context=ctx.obj.data_context,
        event=begin_event_name,
        success=True,
    )
    ctx.obj.usage_event_end = end_event_name


@checkpoint.command(name="new")
@click.argument("name")
@click.option(
    "--jupyter/--no-jupyter",
    is_flag=True,
    help="By default launch jupyter notebooks unless you specify the --no-jupyter flag",
    default=True,
)
@click.pass_context
def checkpoint_new(ctx: click.Context, name: str, jupyter: bool) -> None:
    """Create a new Checkpoint for easy deployments.

    NAME is the name of the Checkpoint to create.
    """
    _checkpoint_new(ctx=ctx, checkpoint_name=name, jupyter=jupyter)


def _checkpoint_new(ctx: click.Context, checkpoint_name: str, jupyter: bool) -> None:

    context: DataContext = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end

    try:
        _verify_checkpoint_does_not_exist(context, checkpoint_name, usage_event_end)

        # Create notebook on disk
        notebook_name = f"edit_checkpoint_{checkpoint_name}.ipynb"
        notebook_file_path = _get_notebook_path(context, notebook_name)
        checkpoint_new_notebook_renderer = CheckpointNewNotebookRenderer(
            context=context, checkpoint_name=checkpoint_name
        )
        checkpoint_new_notebook_renderer.render_to_disk(
            notebook_file_path=notebook_file_path
        )

        if not jupyter:
            cli_message(
                f"To continue editing this Checkpoint, run <green>jupyter notebook {notebook_file_path}</green>"
            )

        send_usage_message(
            data_context=context,
            event=usage_event_end,
            success=True,
        )

        if jupyter:
            cli_message(
                """<green>Because you requested to create a new Checkpoint, we'll open a notebook for you now to edit it!
If you wish to avoid this you can add the `--no-jupyter` flag.</green>\n\n"""
            )
            toolkit.launch_jupyter_notebook(notebook_file_path)

    except Exception as e:
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message=f"<red>{e}</red>",
        )
        return


def _verify_checkpoint_does_not_exist(
    context: DataContext, checkpoint_name: str, usage_event: str
) -> None:
    try:
        if checkpoint_name in context.list_checkpoints():
            toolkit.exit_with_failure_message_and_stats(
                data_context=context,
                usage_event=usage_event,
                message=f"A Checkpoint named `{checkpoint_name}` already exists. Please choose a new name.",
            )
    except InvalidTopLevelConfigKeyError as e:
        toolkit.exit_with_failure_message_and_stats(
            data_context=context, usage_event=usage_event, message=f"<red>{e}</red>"
        )


def _get_notebook_path(context: DataContext, notebook_name: str) -> str:
    return os.path.abspath(
        os.path.join(
            context.root_directory, context.GX_EDIT_NOTEBOOK_DIR, notebook_name
        )
    )


@checkpoint.command(name="list")
@click.pass_context
def checkpoint_list(ctx: click.Context) -> None:
    """List configured checkpoints."""
    context: DataContext = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end

    checkpoints: List[str] = context.list_checkpoints()
    if not checkpoints:
        cli_message(
            "No Checkpoints found.\n"
            "  - Use the command `great_expectations checkpoint new` to create one."
        )
        send_usage_message(
            data_context=context,
            event=usage_event_end,
            success=True,
        )
        sys.exit(0)

    number_found: int = len(checkpoints)
    plural: str = "s" if number_found > 1 else ""
    message: str = f"Found {number_found} Checkpoint{plural}."
    pretty_list: list = [f" - <cyan>{cp}</cyan>" for cp in checkpoints]
    cli_message_list(pretty_list, list_intro_string=message)
    send_usage_message(
        data_context=context,
        event=usage_event_end,
        success=True,
    )


@checkpoint.command(name="delete")
@click.argument("checkpoint")
@click.pass_context
def checkpoint_delete(ctx: click.Context, checkpoint: str) -> None:
    """Delete a Checkpoint."""
    context: DataContext = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end

    try:
        toolkit.delete_checkpoint(
            context=context,
            checkpoint_name=checkpoint,
            usage_event=usage_event_end,
            assume_yes=ctx.obj.assume_yes,
        )
        send_usage_message(
            data_context=context,
            event=usage_event_end,
            success=True,
        )
    except Exception as e:
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message=f"<red>{e}</red>",
        )
        return

    cli_message(f'Checkpoint "{checkpoint}" deleted.')
    sys.exit(0)


@checkpoint.command(name="run")
@click.argument("checkpoint")
@click.pass_context
def checkpoint_run(ctx: click.Context, checkpoint: str) -> None:
    """Run a Checkpoint."""
    context: DataContext = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end

    try:
        result: CheckpointResult = toolkit.run_checkpoint(
            context=context,
            checkpoint_name=checkpoint,
            usage_event=usage_event_end,
        )
    except Exception as e:
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message=f"<red>{e}</red>",
        )
        return

    if not result["success"]:
        cli_message(string="Validation failed!")
        send_usage_message(
            data_context=context,
            event=usage_event_end,
            success=True,
        )
        print_validation_operator_results_details(result=result)
        sys.exit(1)

    cli_message("Validation succeeded!")
    send_usage_message(
        data_context=context,
        event=usage_event_end,
        success=True,
    )
    print_validation_operator_results_details(result=result)
    sys.exit(0)


def print_validation_operator_results_details(
    result: CheckpointResult,
) -> None:
    max_suite_display_width = 40
    cli_message(
        f"""
{'Suite Name'.ljust(max_suite_display_width)}     Status     Expectations met"""
    )
    for result_id, result_item in result.run_results.items():
        vr = result_item["validation_result"]
        stats = vr.statistics
        passed = stats["successful_expectations"]
        evaluated = stats["evaluated_expectations"]
        percentage_slug = (
            f"{round(passed / evaluated * 100, 2) if evaluated > 0 else 100} %"
        )
        stats_slug = f"{passed} of {evaluated} ({percentage_slug})"
        if vr.success:
            status_slug = "<green>✔ Passed</green>"
        else:
            status_slug = "<red>✖ Failed</red>"
        suite_name: str = str(vr.meta["expectation_suite_name"])
        if len(suite_name) > max_suite_display_width:
            suite_name = suite_name[0:max_suite_display_width]
            suite_name = f"{suite_name[:-1]}…"
        status_line: str = f"- {suite_name.ljust(max_suite_display_width)}   {status_slug}   {stats_slug}"
        cli_message(status_line)


@checkpoint.command(name="script")
@click.argument("checkpoint")
@click.pass_context
def checkpoint_script(ctx: click.Context, checkpoint: str) -> None:
    """
    Create a python script to run a Checkpoint.

    Checkpoints can be run directly without this script using the
    `great_expectations Checkpoint run` command.

    This script is provided for those who wish to run Checkpoints via python.
    """
    context: DataContext = ctx.obj.data_context
    usage_event_end: str = ctx.obj.usage_event_end

    toolkit.validate_checkpoint(
        context=context, checkpoint_name=checkpoint, usage_event=usage_event_end
    )

    script_name: str = f"run_{checkpoint}.py"
    script_path: str = os.path.join(
        context.root_directory, context.GX_UNCOMMITTED_DIR, script_name
    )

    if os.path.isfile(script_path):
        toolkit.exit_with_failure_message_and_stats(
            data_context=context,
            usage_event=usage_event_end,
            message=f"""<red>Warning! A script named {script_name} already exists and this command will not overwrite it.</red>
  - Existing file path: {script_path}""",
        )

    _write_checkpoint_script_to_disk(
        context_directory=context.root_directory,
        checkpoint_name=checkpoint,
        script_path=script_path,
    )
    cli_message(
        f"""<green>A python script was created that runs the Checkpoint named: `{checkpoint}`</green>
  - The script is located in `great_expectations/uncommitted/run_{checkpoint}.py`
  - The script can be run with `python great_expectations/uncommitted/run_{checkpoint}.py`"""
    )
    send_usage_message(
        data_context=context,
        event=usage_event_end,
        success=True,
    )


def _write_checkpoint_script_to_disk(
    context_directory: str, checkpoint_name: str, script_path: str
) -> None:
    script_full_path: str = os.path.abspath(os.path.join(script_path))
    template: str = _load_script_template().format(checkpoint_name, context_directory)
    linted_code: str = lint_code(code=template)
    with open(script_full_path, "w") as f:
        f.write(linted_code)


def _load_script_template() -> str:
    with open(file_relative_path(__file__, "checkpoint_script_template.py")) as f:
        template = f.read()
    return template



from great_expectations.core import ExpectationSuiteValidationResult, RunIdentifier
from great_expectations.data_context.types.resource_identifiers import (
    BatchIdentifier,
    ExpectationSuiteIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.render.renderer import MicrosoftTeamsRenderer


def test_MicrosoftTeams_validation_results_with_datadocs():
    validation_result_suite = ExpectationSuiteValidationResult(
        results=[],
        success=True,
        statistics={
            "evaluated_expectations": 0,
            "successful_expectations": 0,
            "unsuccessful_expectations": 0,
            "success_percent": None,
        },
        meta={
            "great_expectations_version": "v0.8.0__develop",
            "expectation_suite_name": "asset.default",
            "run_id": "test_100",
        },
    )

    validation_result_suite_identifier = ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier("asset.default"),
        run_id=RunIdentifier(
            run_name="test_100", run_time="Tue May 08 15:14:45 +0800 2012"
        ),
        batch_identifier=BatchIdentifier(
            batch_identifier="1234", data_asset_name="asset"
        ),
    )

    data_docs_pages = {"local_site": "file:///localsite/index.html"}

    rendered_output = MicrosoftTeamsRenderer().render(
        validation_result_suite, validation_result_suite_identifier, data_docs_pages
    )

    expected_output = {
        "attachments": [
            {
                "content": {
                    "$schema": "http://adaptivecards.io/schemas/adaptive-card.json",
                    "actions": [
                        {
                            "title": "Open data docs",
                            "type": "Action.OpenUrl",
                            "url": "file:///localsite/index.html",
                        }
                    ],
                    "body": [
                        {
                            "height": "auto",
                            "items": [
                                {
                                    "columns": [
                                        {
                                            "items": [
                                                {
                                                    "size": "large",
                                                    "text": "Validation " "results",
                                                    "type": "TextBlock",
                                                    "weight": "bolder",
                                                    "wrap": "true",
                                                },
                                                {
                                                    "isSubtle": "true",
                                                    "spacing": "none",
                                                    "text": "May "
                                                    "08 "
                                                    "2012 "
                                                    "07:14:45",
                                                    "type": "TextBlock",
                                                    "wrap": "true",
                                                },
                                            ],
                                            "type": "Column",
                                            "width": "stretch",
                                        }
                                    ],
                                    "type": "ColumnSet",
                                }
                            ],
                            "separator": "true",
                            "type": "Container",
                        },
                        {
                            "height": "auto",
                            "items": [
                                {
                                    "color": "good",
                                    "horizontalAlignment": "left",
                                    "text": "**Batch validation "
                                    "status:** Success "
                                    "!!!",
                                    "type": "TextBlock",
                                },
                                {
                                    "horizontalAlignment": "left",
                                    "text": "**Data asset " "name:** asset",
                                    "type": "TextBlock",
                                },
                                {
                                    "horizontalAlignment": "left",
                                    "text": "**Expectation "
                                    "suite name:** "
                                    "asset.default",
                                    "type": "TextBlock",
                                },
                                {
                                    "horizontalAlignment": "left",
                                    "text": "**Run name:** " "test_100",
                                    "type": "TextBlock",
                                },
                                {
                                    "horizontalAlignment": "left",
                                    "text": "**Batch ID:** 1234",
                                    "type": "TextBlock",
                                },
                                {
                                    "horizontalAlignment": "left",
                                    "text": "**Summary:** *0* "
                                    "of *0* "
                                    "expectations were "
                                    "met",
                                    "type": "TextBlock",
                                },
                            ],
                            "separator": "true",
                            "type": "Container",
                        },
                    ],
                    "type": "AdaptiveCard",
                    "version": "1.0",
                },
                "contentType": "application/vnd.microsoft.card.adaptive",
            }
        ],
        "type": "message",
    }

    assert rendered_output == expected_output



from __future__ import annotations

import copy
import datetime
import hashlib
import itertools
import logging
import re
import uuid
import warnings
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import scipy.stats as stats
from typing_extensions import Protocol, TypeGuard

import great_expectations.exceptions as gx_exceptions
from great_expectations.core import ExpectationSuite
from great_expectations.core.batch import (
    Batch,
    BatchRequest,
    BatchRequestBase,
    RuntimeBatchRequest,
    materialize_batch_request,
)
from great_expectations.core.domain import (
    INFERRED_SEMANTIC_TYPE_KEY,
    SemanticDomainTypes,
)
from great_expectations.core.metric_domain_types import (
    MetricDomainTypes,  # noqa: TCH001
)
from great_expectations.rule_based_profiler.estimators.numeric_range_estimation_result import (
    NUM_HISTOGRAM_BINS,
    NumericRangeEstimationResult,
)
from great_expectations.rule_based_profiler.parameter_container import (
    FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER,
    VARIABLES_PREFIX,
    Domain,
    ParameterContainer,
    ParameterNode,
    get_parameter_value_by_fully_qualified_parameter_name,
    is_fully_qualified_parameter_name_literal_string_format,
)
from great_expectations.types import safe_deep_copy
from great_expectations.util import (
    convert_ndarray_datetime_to_float_dtype_utc_timezone,
    convert_ndarray_float_to_datetime_dtype,
    convert_ndarray_to_datetime_dtype_best_effort,
    numpy_quantile,
)
from great_expectations.validator.computed_metric import MetricValue  # noqa: TCH001
from great_expectations.validator.metric_configuration import (
    MetricConfiguration,  # noqa: TCH001
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )
    from great_expectations.validator.validator import Validator

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

NP_EPSILON: Union[Number, np.float64] = np.finfo(float).eps

TEMPORARY_EXPECTATION_SUITE_NAME_PREFIX: str = "tmp"
TEMPORARY_EXPECTATION_SUITE_NAME_STEM: str = "suite"
TEMPORARY_EXPECTATION_SUITE_NAME_PATTERN: re.Pattern = re.compile(
    rf"^{TEMPORARY_EXPECTATION_SUITE_NAME_PREFIX}\..+\.{TEMPORARY_EXPECTATION_SUITE_NAME_STEM}\.\w{8}"
)

RECOGNIZED_QUANTILE_STATISTIC_INTERPOLATION_METHODS: set = {
    "auto",
    "nearest",
    "linear",
}


def get_validator(
    purpose: str,
    *,
    data_context: Optional[AbstractDataContext] = None,
    batch_list: Optional[List[Batch]] = None,
    batch_request: Optional[Union[str, BatchRequestBase, dict]] = None,
    domain: Optional[Domain] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> Optional[Validator]:
    validator: Optional[Validator]

    expectation_suite_name: str = f"tmp.{purpose}"
    if domain is None:
        expectation_suite_name = (
            f"{expectation_suite_name}_suite_{str(uuid.uuid4())[:8]}"
        )
    else:
        expectation_suite_name = (
            f"{expectation_suite_name}_{domain.id}_suite_{str(uuid.uuid4())[:8]}"
        )

    batch: Batch
    if batch_list is None or all([batch is None for batch in batch_list]):
        if batch_request is None:
            return None

        batch_request = build_batch_request(
            domain=domain,
            batch_request=batch_request,
            variables=variables,
            parameters=parameters,
        )
    else:
        num_batches: int = len(batch_list)
        if num_batches == 0:
            raise gx_exceptions.ProfilerExecutionError(
                message=f"""{__name__}.get_validator() must utilize at least one Batch ({num_batches} are available).
"""
            )

    validator = get_validator_with_expectation_suite(
        data_context=data_context,
        batch_list=batch_list,
        batch_request=batch_request,
        expectation_suite=None,
        expectation_suite_name=expectation_suite_name,
        component_name=f"rule_based_profiler-{expectation_suite_name}",
        persist=False,
    )

    # Always disabled for RBP and DataAssistants due to volume of metric calculations
    validator.show_progress_bars = False

    return validator


def get_batch_ids(
    data_context: Optional[AbstractDataContext] = None,
    batch_list: Optional[List[Batch]] = None,
    batch_request: Optional[Union[str, BatchRequestBase, dict]] = None,
    limit: Optional[int] = None,
    domain: Optional[Domain] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> Optional[List[str]]:
    batch: Batch
    if batch_list is None or all([batch is None for batch in batch_list]):
        if batch_request is None:
            return None

        batch_request = build_batch_request(
            domain=domain,
            batch_request=batch_request,
            variables=variables,
            parameters=parameters,
        )

        batch_list = data_context.get_batch_list(batch_request=batch_request)

    batch_ids: List[str] = [batch.id for batch in batch_list]

    num_batch_ids: int = len(batch_ids)

    if limit is not None:
        # No need to verify that type of "limit" is "integer", because static type checking already ascertains this.
        if not (0 <= limit <= num_batch_ids):
            raise gx_exceptions.ProfilerExecutionError(
                message=f"""{__name__}.get_batch_ids() allows integer limit values between 0 and {num_batch_ids} \
({limit} was requested).
"""
            )
        batch_ids = batch_ids[-limit:]

    if num_batch_ids == 0:
        raise gx_exceptions.ProfilerExecutionError(
            message=f"""{__name__}.get_batch_ids() must return at least one batch_id ({num_batch_ids} were retrieved).
"""
        )

    return batch_ids


def build_batch_request(
    batch_request: Optional[Union[str, BatchRequestBase, dict]] = None,
    domain: Optional[Domain] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> Optional[Union[BatchRequest, RuntimeBatchRequest]]:
    if batch_request is None:
        return None

    # Obtain BatchRequest from "rule state" (i.e., variables and parameters); from instance variable otherwise.
    effective_batch_request: Optional[
        Union[BatchRequestBase, dict]
    ] = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=batch_request,
        expected_return_type=(BatchRequestBase, dict),
        variables=variables,
        parameters=parameters,
    )
    materialized_batch_request: Optional[
        Union[BatchRequest, RuntimeBatchRequest]
    ] = materialize_batch_request(batch_request=effective_batch_request)

    return materialized_batch_request


def build_metric_domain_kwargs(
    batch_id: Optional[str] = None,
    metric_domain_kwargs: Optional[Union[str, dict]] = None,
    domain: Optional[Domain] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
):
    # Obtain domain kwargs from "rule state" (i.e., variables and parameters); from instance variable otherwise.
    metric_domain_kwargs = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=metric_domain_kwargs,
        expected_return_type=None,
        variables=variables,
        parameters=parameters,
    )
    if metric_domain_kwargs is None:
        metric_domain_kwargs = {}

    metric_domain_kwargs = copy.deepcopy(metric_domain_kwargs)

    if batch_id:
        metric_domain_kwargs["batch_id"] = batch_id

    return metric_domain_kwargs


def get_parameter_value_and_validate_return_type(
    domain: Optional[Domain] = None,
    parameter_reference: Optional[Union[Any, str]] = None,
    expected_return_type: Optional[Union[type, tuple]] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> Any:
    """
    This method allows for the parameter_reference to be specified as an object (literal, dict, any typed object, etc.)
    or as a fully-qualified parameter name.  In either case, it can optionally validate the type of the return value.
    """
    if isinstance(parameter_reference, dict):
        parameter_reference = safe_deep_copy(data=parameter_reference)

    parameter_reference = get_parameter_value(
        domain=domain,
        parameter_reference=parameter_reference,
        variables=variables,
        parameters=parameters,
    )

    if expected_return_type is not None:
        if not isinstance(parameter_reference, expected_return_type):
            raise gx_exceptions.ProfilerExecutionError(
                message=f"""Argument "{parameter_reference}" must be of type "{str(expected_return_type)}" \
(value of type "{str(type(parameter_reference))}" was encountered).
"""
            )

    return parameter_reference


def get_parameter_value(
    domain: Optional[Domain] = None,
    parameter_reference: Optional[Union[Any, str]] = None,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> Optional[Any]:
    """
    This method allows for the parameter_reference to be specified as an object (literal, dict, any typed object, etc.)
    or as a fully-qualified parameter name.  Moreover, if the parameter_reference argument is an object of type "dict",
    it will recursively detect values using the fully-qualified parameter name format and evaluate them accordingly.
    """
    if isinstance(parameter_reference, dict):
        for key, value in parameter_reference.items():
            parameter_reference[key] = get_parameter_value(
                domain=domain,
                parameter_reference=value,
                variables=variables,
                parameters=parameters,
            )
    elif isinstance(parameter_reference, (list, set, tuple)):
        parameter_reference_type: type = type(parameter_reference)
        element: Any
        return parameter_reference_type(
            [
                get_parameter_value(
                    domain=domain,
                    parameter_reference=element,
                    variables=variables,
                    parameters=parameters,
                )
                for element in parameter_reference
            ]
        )
    elif isinstance(
        parameter_reference, str
    ) and is_fully_qualified_parameter_name_literal_string_format(
        fully_qualified_parameter_name=parameter_reference
    ):
        parameter_reference = get_parameter_value_by_fully_qualified_parameter_name(
            fully_qualified_parameter_name=parameter_reference,
            domain=domain,
            variables=variables,
            parameters=parameters,
        )
        parameter_reference = get_parameter_value(
            domain=domain,
            parameter_reference=parameter_reference,
            variables=variables,
            parameters=parameters,
        )

    return parameter_reference


def get_resolved_metrics_by_key(
    validator: Validator,
    metric_configurations_by_key: Dict[str, List[MetricConfiguration]],
    runtime_configuration: Optional[dict] = None,
) -> Dict[str, Dict[Tuple[str, str, str], MetricValue]]:
    """
    Compute (resolve) metrics for every column name supplied on input.

    Args:
        validator: Validator used to compute column cardinality.
        metric_configurations_by_key: metric configurations used to compute figures of merit.
        Dictionary of the form {
            "my_key": List[MetricConfiguration],  # examples of "my_key" are: "my_column_name", "my_batch_id", etc.
        }
        runtime_configuration: Additional run-time settings (see "Validator.DEFAULT_RUNTIME_CONFIGURATION").

    Returns:
        Dictionary of the form {
            "my_key": Dict[Tuple[str, str, str], MetricValue],
        }
    """
    key: str
    metric_configuration: MetricConfiguration
    metric_configurations_for_key: List[MetricConfiguration]

    # Step 1: Gather "MetricConfiguration" objects corresponding to all possible key values/combinations.
    # and compute all metric values (resolve "MetricConfiguration" objects ) using a single method call.
    resolved_metrics: Dict[
        Tuple[str, str, str], MetricValue
    ] = validator.compute_metrics(
        metric_configurations=[
            metric_configuration
            for key, metric_configurations_for_key in metric_configurations_by_key.items()
            for metric_configuration in metric_configurations_for_key
        ],
        runtime_configuration=runtime_configuration,
        min_graph_edges_pbar_enable=0,
    )

    # Step 2: Gather "MetricConfiguration" ID values for each key (one element per batch_id in every list).
    metric_configuration_ids_by_key: Dict[str, List[Tuple[str, str, str]]] = {
        key: [
            metric_configuration.id
            for metric_configuration in metric_configurations_for_key
        ]
        for key, metric_configurations_for_key in metric_configurations_by_key.items()
    }

    metric_configuration_ids: List[Tuple[str, str, str]]
    # Step 3: Obtain flattened list of "MetricConfiguration" ID values across all key values/combinations.
    metric_configuration_ids_all_keys: List[Tuple[str, str, str]] = list(
        itertools.chain(
            *[
                metric_configuration_ids
                for metric_configuration_ids in metric_configuration_ids_by_key.values()
            ]
        )
    )

    # Step 4: Retain only those metric computation results that both, correspond to "MetricConfiguration" objects of
    # interest (reflecting specified key values/combinations).
    metric_configuration_id: Tuple[str, str, str]
    metric_value: Any
    resolved_metrics = {
        metric_configuration_id: metric_value
        for metric_configuration_id, metric_value in resolved_metrics.items()
        if metric_configuration_id in metric_configuration_ids_all_keys
    }

    # Step 5: Gather "MetricConfiguration" ID values for effective collection of resolved metrics.
    metric_configuration_ids_resolved_metrics: List[Tuple[str, str, str]] = list(
        resolved_metrics.keys()
    )

    # Step 6: Produce "key" list, corresponding to effective "MetricConfiguration" ID values.
    candidate_keys: List[str] = [
        key
        for key, metric_configuration_ids in metric_configuration_ids_by_key.items()
        if all(
            [
                metric_configuration_id in metric_configuration_ids_resolved_metrics
                for metric_configuration_id in metric_configuration_ids
            ]
        )
    ]

    resolved_metrics_by_key: Dict[str, Dict[Tuple[str, str, str], MetricValue]] = {
        key: {
            metric_configuration.id: resolved_metrics[metric_configuration.id]
            for metric_configuration in metric_configurations_by_key[key]
        }
        for key in candidate_keys
    }

    return resolved_metrics_by_key


def build_domains_from_column_names(
    rule_name: str,
    column_names: List[str],
    domain_type: MetricDomainTypes,
    table_column_name_to_inferred_semantic_domain_type_map: Optional[
        Dict[str, SemanticDomainTypes]
    ] = None,
) -> List[Domain]:
    """
    This utility method builds "simple" Domain objects (i.e., required fields only, no "details" metadata accepted).

    :param rule_name: name of Rule object, for which "Domain" objects are obtained.
    :param column_names: list of column names to serve as values for "column" keys in "domain_kwargs" dictionary
    :param domain_type: type of Domain objects (same "domain_type" must be applicable to all Domain objects returned)
    :param table_column_name_to_inferred_semantic_domain_type_map: map from column name to inferred semantic type
    :return: list of resulting Domain objects
    """
    column_name: str
    domains: List[Domain] = [
        Domain(
            domain_type=domain_type,
            domain_kwargs={
                "column": column_name,
            },
            details={
                INFERRED_SEMANTIC_TYPE_KEY: {
                    column_name: table_column_name_to_inferred_semantic_domain_type_map[
                        column_name
                    ],
                }
                if table_column_name_to_inferred_semantic_domain_type_map
                else None,
            },
            rule_name=rule_name,
        )
        for column_name in column_names
    ]

    return domains


def convert_variables_to_dict(
    variables: Optional[ParameterContainer] = None,
) -> Dict[str, Any]:
    variables_as_dict: Optional[
        Union[ParameterNode, Dict[str, Any]]
    ] = get_parameter_value_and_validate_return_type(
        domain=None,
        parameter_reference=VARIABLES_PREFIX,
        expected_return_type=None,
        variables=variables,
        parameters=None,
    )
    if isinstance(variables_as_dict, ParameterNode):
        return variables_as_dict.to_dict()

    if variables_as_dict is None:
        return {}

    return variables_as_dict


def integer_semantic_domain_type(domain: Domain) -> bool:
    """
    This method examines "INFERRED_SEMANTIC_TYPE_KEY" attribute of "Domain" argument to check whether or not underlying
    "SemanticDomainTypes" enum value is an "integer".  Because explicitly designated "SemanticDomainTypes.INTEGER" type
    is unavaiable, "SemanticDomainTypes.LOGIC" and "SemanticDomainTypes.IDENTIFIER" are intepreted as "integer" values.

    This method can be used "NumericMetricRangeMultiBatchParameterBuilder._get_round_decimals_using_heuristics()".

    Note: Inability to assess underlying "SemanticDomainTypes" details of "Domain" object produces "False" return value.

    Args:
        domain: "Domain" object to inspect for underlying "SemanticDomainTypes" details

    Returns:
        Boolean value indicating whether or not specified "Domain" is inferred to denote "integer" values

    """

    inferred_semantic_domain_type: Dict[str, SemanticDomainTypes] = domain.details.get(
        INFERRED_SEMANTIC_TYPE_KEY
    )

    semantic_domain_type: SemanticDomainTypes
    return inferred_semantic_domain_type and all(
        [
            semantic_domain_type
            in [
                SemanticDomainTypes.LOGIC,
                SemanticDomainTypes.IDENTIFIER,
            ]
            for semantic_domain_type in inferred_semantic_domain_type.values()
        ]
    )


def datetime_semantic_domain_type(domain: Domain) -> bool:
    """
    This method examines "INFERRED_SEMANTIC_TYPE_KEY" attribute of "Domain" argument to check whether or not underlying
    "SemanticDomainTypes" enum value is "SemanticDomainTypes.DATETIME".

    Note: Inability to assess underlying "SemanticDomainTypes" details of "Domain" object produces "False" return value.

    Args:
        domain: "Domain" object to inspect for underlying "SemanticDomainTypes" details

    Returns:
        Boolean value indicating whether or not specified "Domain" is inferred as "SemanticDomainTypes.DATETIME"
    """

    inferred_semantic_domain_type: Dict[str, SemanticDomainTypes] = domain.details.get(
        INFERRED_SEMANTIC_TYPE_KEY
    )

    semantic_domain_type: SemanticDomainTypes
    return inferred_semantic_domain_type and all(
        [
            semantic_domain_type == SemanticDomainTypes.DATETIME
            for semantic_domain_type in inferred_semantic_domain_type.values()
        ]
    )


def get_false_positive_rate_from_rule_state(
    false_positive_rate: Union[str, float],
    domain: Domain,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> Union[float, np.float64]:
    """
    This method obtains false_positive_rate from "rule state" (i.e., variables and parameters) and validates the result.
    """
    if false_positive_rate is None:
        return 5.0e-2

    # Obtain false_positive_rate from "rule state" (i.e., variables and parameters); from instance variable otherwise.
    false_positive_rate = get_parameter_value_and_validate_return_type(
        domain=domain,
        parameter_reference=false_positive_rate,
        expected_return_type=(float, np.float64),
        variables=variables,
        parameters=parameters,
    )
    if not (0.0 <= false_positive_rate <= 1.0):
        raise gx_exceptions.ProfilerExecutionError(
            f"""false_positive_rate must be a positive decimal number between 0 and 1 inclusive [0, 1], but \
{false_positive_rate} was provided.
"""
        )
    elif false_positive_rate <= NP_EPSILON:
        warnings.warn(
            f"""You have chosen a false_positive_rate of {false_positive_rate}, which is too close to 0.  A \
false_positive_rate of {NP_EPSILON} has been selected instead.
"""
        )
        false_positive_rate = np.float64(NP_EPSILON)
    elif false_positive_rate >= (1.0 - NP_EPSILON):
        warnings.warn(
            f"""You have chosen a false_positive_rate of {false_positive_rate}, which is too close to 1.  A \
false_positive_rate of {1.0 - NP_EPSILON} has been selected instead.
"""
        )
        false_positive_rate = np.float64(1.0 - NP_EPSILON)

    return false_positive_rate


def get_quantile_statistic_interpolation_method_from_rule_state(
    quantile_statistic_interpolation_method: str,
    round_decimals: int,
    domain: Domain,
    variables: Optional[ParameterContainer] = None,
    parameters: Optional[Dict[str, ParameterContainer]] = None,
) -> str:
    """
    This method obtains quantile_statistic_interpolation_method from "rule state" (i.e., variables and parameters) and
    validates the result.
    """
    # Obtain quantile_statistic_interpolation_method directive from "rule state" (i.e., variables and parameters); from instance variable otherwise.
    quantile_statistic_interpolation_method = (
        get_parameter_value_and_validate_return_type(
            domain=domain,
            parameter_reference=quantile_statistic_interpolation_method,
            expected_return_type=str,
            variables=variables,
            parameters=parameters,
        )
    )
    if (
        quantile_statistic_interpolation_method
        not in RECOGNIZED_QUANTILE_STATISTIC_INTERPOLATION_METHODS
    ):
        raise gx_exceptions.ProfilerExecutionError(
            message=f"""The directive "quantile_statistic_interpolation_method" can be only one of \
{RECOGNIZED_QUANTILE_STATISTIC_INTERPOLATION_METHODS} ("{quantile_statistic_interpolation_method}" was detected).
"""
        )

    if quantile_statistic_interpolation_method == "auto":
        if round_decimals == 0:
            quantile_statistic_interpolation_method = "nearest"
        else:
            quantile_statistic_interpolation_method = "linear"

    return quantile_statistic_interpolation_method


def compute_quantiles(
    metric_values: np.ndarray,
    false_positive_rate: np.float64,
    quantile_statistic_interpolation_method: str,
) -> NumericRangeEstimationResult:
    lower_quantile = numpy_quantile(
        a=metric_values,
        q=(false_positive_rate / 2.0),
        axis=0,
        method=quantile_statistic_interpolation_method,
    )
    upper_quantile = numpy_quantile(
        a=metric_values,
        q=1.0 - (false_positive_rate / 2.0),
        axis=0,
        method=quantile_statistic_interpolation_method,
    )
    return build_numeric_range_estimation_result(
        metric_values=metric_values,
        min_value=lower_quantile,
        max_value=upper_quantile,
    )


def compute_kde_quantiles_point_estimate(
    metric_values: np.ndarray,
    false_positive_rate: np.float64,
    n_resamples: int,
    quantile_statistic_interpolation_method: str,
    bw_method: Union[str, float, Callable],
    random_seed: Optional[int] = None,
) -> NumericRangeEstimationResult:
    """
    ML Flow Experiment: parameter_builders_bootstrap/kde_quantiles
    ML Flow Experiment ID: 721280826919117

    An internal implementation of the "kernel density estimation" estimator method, returning a point estimate for a
    population parameter of interest (lower and upper quantiles in this case).

    Overview: https://en.wikipedia.org/wiki/Kernel_density_estimation
    Bandwidth Effect: https://en.wikipedia.org/wiki/Kernel_density_estimation#Bandwidth_selection
    Bandwidth Method: https://stats.stackexchange.com/questions/90656/kernel-bandwidth-scotts-vs-silvermans-rules

    Args:
        metric_values: "numpy.ndarray" of "dtype.float" values with elements corresponding to "Batch" data samples.
        false_positive_rate: user-configured fraction between 0 and 1 expressing desired false positive rate for
            identifying unexpected values as judged by the upper- and lower- quantiles of the observed metric data.
        n_resamples: A positive integer indicating the sample size resulting from the sampling with replacement
            procedure.
        quantile_statistic_interpolation_method: Supplies value of (interpolation) "method" to "np.quantile()"
            statistic, used for confidence intervals.
        bw_method: The estimator bandwidth as described in:
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.gaussian_kde.html
        random_seed: An optional random_seed to pass to "np.random.Generator(np.random.PCG64(random_seed))"
            for making probabilistic sampling deterministic.
    """
    lower_quantile_pct: float = false_positive_rate / 2.0
    upper_quantile_pct: float = 1.0 - (false_positive_rate / 2.0)

    metric_values_density_estimate: stats.gaussian_kde = stats.gaussian_kde(
        metric_values, bw_method=bw_method
    )

    metric_values_gaussian_sample: np.ndarray
    if random_seed:
        metric_values_gaussian_sample = metric_values_density_estimate.resample(
            n_resamples,
            seed=random_seed,
        )
    else:
        metric_values_gaussian_sample = metric_values_density_estimate.resample(
            n_resamples,
        )

    lower_quantile_point_estimate: Union[
        np.float64, datetime.datetime
    ] = numpy_quantile(
        metric_values_gaussian_sample,
        q=lower_quantile_pct,
        method=quantile_statistic_interpolation_method,
    )
    upper_quantile_point_estimate: Union[
        np.float64, datetime.datetime
    ] = numpy_quantile(
        metric_values_gaussian_sample,
        q=upper_quantile_pct,
        method=quantile_statistic_interpolation_method,
    )

    return build_numeric_range_estimation_result(
        metric_values=metric_values,
        min_value=lower_quantile_point_estimate,
        max_value=upper_quantile_point_estimate,
    )


def compute_bootstrap_quantiles_point_estimate(
    metric_values: np.ndarray,
    false_positive_rate: np.float64,
    n_resamples: int,
    quantile_statistic_interpolation_method: str,
    quantile_bias_correction: bool,
    quantile_bias_std_error_ratio_threshold: float,
    random_seed: Optional[int] = None,
) -> NumericRangeEstimationResult:
    """
    ML Flow Experiment: parameter_builders_bootstrap/bootstrap_quantiles
    ML Flow Experiment ID: 4129654509298109

    An internal implementation of the "bootstrap" estimator method, returning a point estimate for a population
    parameter of interest (lower and upper quantiles in this case). See
    https://en.wikipedia.org/wiki/Bootstrapping_(statistics) for an introduction to "bootstrapping" in statistics.

    The methods implemented here can be found in:
    Efron, B., & Tibshirani, R. J. (1993). Estimates of bias. An Introduction to the Bootstrap (pp. 124-130).
        Springer Science and Business Media Dordrecht. DOI 10.1007/978-1-4899-4541-9

    This implementation is sub-par compared to the one available from the "SciPy" standard library
    ("https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.bootstrap.html"), in that it does not handle
    multi-dimensional statistics. "scipy.stats.bootstrap" is vectorized, thus having the ability to accept a
    multi-dimensional statistic function and process all dimensions.

    Unfortunately, as of March 4th, 2022, the SciPy implementation has two issues: 1) it only returns a confidence
    interval and not a point estimate for the population parameter of interest, which is what we require for our use
    cases. 2) It can not handle multi-dimensional statistics and correct for bias simultaneously. You must either use
    one feature or the other.

    This implementation could only be replaced by "scipy.stats.bootstrap" if Great Expectations drops support for
    Python 3.6, thereby enabling us to use a more up-to-date version of the "scipy" Python package (the currently used
    version does not have "bootstrap"). Also, as discussed above, two contributions would need to be made to the SciPy
    package to enable 1) bias correction for multi-dimensional statistics and 2) a return value of a point estimate for
    the population parameter of interest (lower and upper quantiles in this case).

    Additional future direction could include developing enhancements to bootstrapped estimator based on theory
    presented in "http://dido.econ.yale.edu/~dwka/pub/p1001.pdf":
    @article{Andrews2000a,
        added-at = {2008-04-25T10:38:44.000+0200},
        author = {Andrews, Donald W. K. and Buchinsky, Moshe},
        biburl = {https://www.bibsonomy.org/bibtex/28e2f0a58cdb95e39659921f989a17bdd/smicha},
        day = 01,
        interhash = {778746398daa9ba63bdd95391f1efd37},
        intrahash = {8e2f0a58cdb95e39659921f989a17bdd},
        journal = {Econometrica},
        keywords = {imported},
        month = Jan,
        note = {doi: 10.1111/1468-0262.00092},
        number = 1,
        pages = {23--51},
        timestamp = {2008-04-25T10:38:52.000+0200},
        title = {A Three-step Method for Choosing the Number of Bootstrap Repetitions},
        url = {http://www.blackwell-synergy.com/doi/abs/10.1111/1468-0262.00092},
        volume = 68,
        year = 2000
    }
    The article outlines a three-step minimax procedure that relies on the Central Limit Theorem (C.L.T.) along with the
    bootstrap sampling technique (see https://en.wikipedia.org/wiki/Bootstrapping_(statistics) for background) for
    computing the stopping criterion, expressed as the optimal number of bootstrap samples, needed to achieve a maximum
    probability that the value of the statistic of interest will be minimally deviating from its actual (ideal) value.
    """
    lower_quantile_pct: float = false_positive_rate / 2.0
    upper_quantile_pct: float = 1.0 - false_positive_rate / 2.0

    sample_lower_quantile: np.ndarray = numpy_quantile(
        a=metric_values,
        q=lower_quantile_pct,
        method=quantile_statistic_interpolation_method,
    )
    sample_upper_quantile: np.ndarray = numpy_quantile(
        a=metric_values,
        q=upper_quantile_pct,
        method=quantile_statistic_interpolation_method,
    )

    bootstraps: np.ndarray
    if random_seed:
        random_state: np.random.Generator = np.random.Generator(
            np.random.PCG64(random_seed)
        )
        bootstraps = random_state.choice(
            metric_values, size=(n_resamples, metric_values.size)
        )
    else:
        bootstraps = np.random.choice(
            metric_values, size=(n_resamples, metric_values.size)
        )

    lower_quantile_bias_corrected_point_estimate: Union[
        np.float64, datetime.datetime
    ] = _determine_quantile_bias_corrected_point_estimate(
        bootstraps=bootstraps,
        quantile_pct=lower_quantile_pct,
        quantile_statistic_interpolation_method=quantile_statistic_interpolation_method,
        quantile_bias_correction=quantile_bias_correction,
        quantile_bias_std_error_ratio_threshold=quantile_bias_std_error_ratio_threshold,
        sample_quantile=sample_lower_quantile,
    )
    upper_quantile_bias_corrected_point_estimate: Union[
        np.float64, datetime.datetime
    ] = _determine_quantile_bias_corrected_point_estimate(
        bootstraps=bootstraps,
        quantile_pct=upper_quantile_pct,
        quantile_statistic_interpolation_method=quantile_statistic_interpolation_method,
        quantile_bias_correction=quantile_bias_correction,
        quantile_bias_std_error_ratio_threshold=quantile_bias_std_error_ratio_threshold,
        sample_quantile=sample_upper_quantile,
    )

    return build_numeric_range_estimation_result(
        metric_values=metric_values,
        min_value=lower_quantile_bias_corrected_point_estimate,
        max_value=upper_quantile_bias_corrected_point_estimate,
    )


def build_numeric_range_estimation_result(
    metric_values: np.ndarray,
    min_value: Number,
    max_value: Number,
) -> NumericRangeEstimationResult:
    """
    Computes histogram of 1-dimensional set of data points and packages it together with value range as returned output.

    Args:
        metric_values: "numpy.ndarray" of "dtype.float" values with elements corresponding to "Batch" data samples.
        min_value: pre-computed supremum of "metric_values" (properly conditioned for output).
        max_value: pre-computed infimum of "metric_values" (properly conditioned for output).

    Returns:
        Structured "NumericRangeEstimationResult" object, containing histogram and value_range attributes.
    """
    ndarray_is_datetime_type: bool
    metric_values_converted: np.ndarray
    (
        ndarray_is_datetime_type,
        metric_values_converted,
    ) = convert_metric_values_to_float_dtype_best_effort(metric_values=metric_values)

    histogram: Tuple[np.ndarray, np.ndarray]
    bin_edges: np.ndarray
    if ndarray_is_datetime_type:
        histogram = np.histogram(a=metric_values_converted, bins=NUM_HISTOGRAM_BINS)
        # Use "UTC" TimeZone normalization in "bin_edges" when "metric_values" consists of "datetime.datetime" objects.
        bin_edges = convert_ndarray_float_to_datetime_dtype(data=histogram[1])
    else:
        histogram = np.histogram(a=metric_values, bins=NUM_HISTOGRAM_BINS)
        bin_edges = histogram[1]

    return NumericRangeEstimationResult(
        estimation_histogram=np.vstack(
            (
                np.pad(
                    array=histogram[0],
                    pad_width=(0, 1),
                    mode="constant",
                    constant_values=0,
                ),
                bin_edges,
            )
        ),
        value_range=np.asarray([min_value, max_value]),
    )


def _determine_quantile_bias_corrected_point_estimate(
    bootstraps: np.ndarray,
    quantile_pct: float,
    quantile_statistic_interpolation_method: str,
    quantile_bias_correction: bool,
    quantile_bias_std_error_ratio_threshold: float,
    sample_quantile: np.ndarray,
) -> np.float64:
    bootstrap_quantiles: Union[np.ndarray, np.float64] = numpy_quantile(
        bootstraps,
        q=quantile_pct,
        axis=1,
        method=quantile_statistic_interpolation_method,
    )
    bootstrap_quantile_point_estimate: np.ndarray = np.mean(bootstrap_quantiles)
    bootstrap_quantile_standard_error: np.ndarray = np.std(bootstrap_quantiles)
    bootstrap_quantile_bias: float = bootstrap_quantile_point_estimate - sample_quantile

    # Bias / Standard Error > 0.25 is a rule of thumb for when to apply bias correction.
    # See:
    # Efron, B., & Tibshirani, R. J. (1993). Estimates of bias. An Introduction to the Bootstrap (pp. 128).
    #         Springer Science and Business Media Dordrecht. DOI 10.1007/978-1-4899-4541-9
    quantile_bias_corrected_point_estimate: np.float64

    if (
        not quantile_bias_correction
        and bootstrap_quantile_standard_error > 0.0
        and bootstrap_quantile_bias / bootstrap_quantile_standard_error
        <= quantile_bias_std_error_ratio_threshold
    ):
        quantile_bias_corrected_point_estimate = bootstrap_quantile_point_estimate
    else:
        quantile_bias_corrected_point_estimate = (
            bootstrap_quantile_point_estimate - bootstrap_quantile_bias
        )

    return quantile_bias_corrected_point_estimate


def convert_metric_values_to_float_dtype_best_effort(
    metric_values: np.ndarray,
) -> Tuple[bool, np.ndarray]:
    """
    Makes best effort attempt to discern element type of 1-D "np.ndarray" and convert it to "float" "np.ndarray" type.

    Note: Conversion of "datetime.datetime" to "float" uses "UTC" TimeZone to normalize all "datetime.datetime" values.

    Return:
        Boolean flag -- True, if conversion of original "np.ndarray" to "datetime.datetime" occurred; False, otherwise.
    """
    original_ndarray_is_datetime_type: bool
    conversion_ndarray_to_datetime_type_performed: bool
    metric_values_converted: np.ndaarray
    (
        original_ndarray_is_datetime_type,
        conversion_ndarray_to_datetime_type_performed,
        metric_values_converted,
    ) = convert_ndarray_to_datetime_dtype_best_effort(
        data=metric_values,
        datetime_detected=False,
        parse_strings_as_datetimes=True,
        fuzzy=False,
    )
    ndarray_is_datetime_type: bool = (
        original_ndarray_is_datetime_type
        or conversion_ndarray_to_datetime_type_performed
    )
    if ndarray_is_datetime_type:
        metric_values_converted = convert_ndarray_datetime_to_float_dtype_utc_timezone(
            data=metric_values_converted
        )
    else:
        metric_values_converted = metric_values

    return ndarray_is_datetime_type, metric_values_converted


def get_validator_with_expectation_suite(
    data_context: AbstractDataContext,
    batch_list: Optional[List[Batch]] = None,
    batch_request: Optional[Union[BatchRequestBase, dict]] = None,
    expectation_suite: Optional[ExpectationSuite] = None,
    expectation_suite_name: Optional[str] = None,
    component_name: str = "test",
    persist: bool = False,
) -> Validator:
    """
    Instantiates and returns "Validator" using "data_context", "batch_list" or "batch_request", and other information.
    Use "expectation_suite" if provided; otherwise, if "expectation_suite_name" is specified, then create
    "ExpectationSuite" from it.  Otherwise, generate temporary "expectation_suite_name" using supplied "component_name".
    """
    assert expectation_suite is None or isinstance(expectation_suite, ExpectationSuite)
    assert expectation_suite_name is None or isinstance(expectation_suite_name, str)

    expectation_suite = get_or_create_expectation_suite(
        data_context=data_context,
        expectation_suite=expectation_suite,
        expectation_suite_name=expectation_suite_name,
        component_name=component_name,
        persist=persist,
    )
    batch_request = materialize_batch_request(batch_request=batch_request)
    validator: Validator = data_context.get_validator(
        batch_list=batch_list,
        batch_request=batch_request,
        expectation_suite=expectation_suite,
    )

    return validator


def get_or_create_expectation_suite(
    data_context: Optional[AbstractDataContext],
    expectation_suite: Optional[ExpectationSuite] = None,
    expectation_suite_name: Optional[str] = None,
    component_name: Optional[str] = None,
    persist: bool = False,
) -> ExpectationSuite:
    """
    Use "expectation_suite" if provided.  If not, then if "expectation_suite_name" is specified, then create
    "ExpectationSuite" from it.  Otherwise, generate temporary "expectation_suite_name" using supplied "component_name".
    """
    generate_temp_expectation_suite_name: bool
    create_expectation_suite: bool

    if expectation_suite is not None and expectation_suite_name is not None:
        if expectation_suite.expectation_suite_name != expectation_suite_name:
            raise ValueError(
                'Mutually inconsistent "expectation_suite" and "expectation_suite_name" were specified.'
            )

        return expectation_suite
    elif expectation_suite is None and expectation_suite_name is not None:
        generate_temp_expectation_suite_name = False
        create_expectation_suite = True
    elif expectation_suite is not None and expectation_suite_name is None:
        generate_temp_expectation_suite_name = False
        create_expectation_suite = False
    else:
        generate_temp_expectation_suite_name = True
        create_expectation_suite = True

    if generate_temp_expectation_suite_name:
        if not component_name:
            component_name = "test"

        expectation_suite_name = f"{TEMPORARY_EXPECTATION_SUITE_NAME_PREFIX}.{component_name}.{TEMPORARY_EXPECTATION_SUITE_NAME_STEM}.{str(uuid.uuid4())[:8]}"

    if create_expectation_suite:
        if persist:
            try:
                # noinspection PyUnusedLocal
                expectation_suite = data_context.get_expectation_suite(
                    expectation_suite_name=expectation_suite_name
                )
            except gx_exceptions.DataContextError:
                expectation_suite = data_context.add_expectation_suite(
                    expectation_suite_name=expectation_suite_name
                )
                logger.info(
                    f'Created ExpectationSuite "{expectation_suite.expectation_suite_name}".'
                )
        else:
            expectation_suite = ExpectationSuite(
                expectation_suite_name=expectation_suite_name,
                data_context=data_context,
            )

    return expectation_suite


def sanitize_parameter_name(
    name: str,
    suffix: Optional[str] = None,
) -> str:
    """
    This method provides display-friendly version of "name" argument (with optional "suffix" argument).

    In most situations, "suffix" is not needed.  However, in certain use cases, "name" argument is same among different
    calling structures, whereby "name" can be disambiguated by addition of distinguishing "suffix" argument.  Using
    list of "MetricConfiguration" objects as example, "metric_name" and "metric_domain_kwargs" are commonly same among
    them, while "metric_value_kwargs" are different (i.e., calculating values of same metric with different parameters).
    In this case, supplying "MetricConfiguration.metric_domain_kwargs_id" as "suffix" makes sanitized names unique.

    Args:
        name: string-valued "name" argument to be sanitized
        suffix: additional component (i.e., "suffix") argument to be appended to "name" and sanitized together

    Returns:
        string-valued sanitized concatenation of "name" and MD5-digest of "suffix" arguments.
    """
    if suffix:
        suffix = hashlib.md5(suffix.encode("utf-8")).hexdigest()
        name = f"{name}{FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER}{suffix}"

    return name.replace(FULLY_QUALIFIED_PARAMETER_NAME_SEPARATOR_CHARACTER, "_")


class _NumericIterableWithDtype(Iterable, Protocol):
    @property
    def dtype(self) -> Any:
        ...


def _is_iterable_of_numeric_dtypes(
    obj: Any,
) -> TypeGuard[_NumericIterableWithDtype]:
    if hasattr(obj, "dtype") and np.issubdtype(obj.dtype, np.number):
        return True
    return False



from __future__ import annotations

import copy
import dataclasses
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Type,
    Union,
    cast,
)

from typing_extensions import Literal

from great_expectations.compatibility import pydantic
import great_expectations.exceptions as gx_exceptions
from great_expectations.core.batch_spec import SqlAlchemyDatasourceBatchSpec
from great_expectations.experimental.datasources.experimental_base_model import (
    ExperimentalBaseModel,
)
from great_expectations.experimental.datasources.interfaces import (
    Batch,
    BatchRequest,
    BatchRequestOptions,
    BatchSorter,
    BatchSortersDefinition,
    DataAsset,
    Datasource,
    TestConnectionError,
)

SQLALCHEMY_IMPORTED = False
try:
    import sqlalchemy

    SQLALCHEMY_IMPORTED = True
except ImportError:
    pass

if TYPE_CHECKING:
    from great_expectations.execution_engine import SqlAlchemyExecutionEngine


class SQLDatasourceError(Exception):
    pass


class SQLDatasourceWarning(UserWarning):
    pass


class _ColumnSplitter(ExperimentalBaseModel):
    column_name: str
    method_name: str

    @property
    def param_names(self) -> List[str]:
        """Returns the parameter names that specify a batch derived from this column splitter

        For example, for ColumnSplitterYearMonth this returns ["year", "month"]. For more
        examples, please see _ColumnSplitter subclasses.
        """
        raise NotImplementedError

    def splitter_method_kwargs(self) -> Dict[str, Any]:
        """A shim to our sqlalchemy execution engine splitter methods

        We translate any internal _ColumnSplitter state and what is passed in from
        a batch_request to the splitter_kwargs required by our execution engine
        data splitters defined in:
        great_expectations.execution_engine.split_and_sample.sqlalchemy_data_splitter

        Look at _ColumnSplitter subclasses for concrete examples.
        """
        raise NotImplementedError

    def batch_batch_request_options_to_batch_spec_kwarg_identifiers(
        self, options: BatchRequestOptions
    ) -> Dict[str, Any]:
        """Translate the batch request options to the dictionary needed to by our execution engine

        Arguments:
            options: A BatchRequest.options dictionary that specified all the fields necessary
                     to specify a batch with respect to this column splitter.

        Returns:
            A dictionary that can be added to batch_spec_kwargs["batch_identifiers"] when we
            make a SqlAlchemyDatasourceBatchSpec and run execution_engine.get_batch_data_and_markers
            for a particular batch specified by options.
        """
        raise NotImplementedError

    def param_defaults(self, sql_asset: _SQLAsset) -> List[Dict]:
        from great_expectations.execution_engine.split_and_sample.sqlalchemy_data_splitter import (
            SqlAlchemyDataSplitter,
        )

        execution_engine = sql_asset.datasource.get_execution_engine()
        splitter = SqlAlchemyDataSplitter(execution_engine.dialect_name)
        batch_identifier_data = splitter.get_data_for_batch_identifiers(
            execution_engine=execution_engine,
            selectable=sql_asset.as_selectable(),
            splitter_method_name=self.method_name,
            splitter_kwargs=self.splitter_method_kwargs(),
        )
        params: List[Dict] = []
        for identifier_data in batch_identifier_data:
            # batch_identifiers is a list of dicts that is keyed by the column name. The value is either:
            #  * A dict whose keys are batch request parameter names and values are the batch values.
            #  * A single value for the 1 batch request parameter.
            value = identifier_data[self.column_name]
            if isinstance(value, dict):
                params.append(value)
            elif len(self.param_names) == 1:
                params.append({self.param_names[0]: value})
            else:
                raise SQLDatasourceError(
                    f"identifier_data is in an unexpected form for {self.__class__} "
                    f"on column {self.column_name}: {identifier_data}"
                )
        return params

    def test_connection(self, table_asset: TableAsset) -> None:
        """Test the connection for the ColumnSplitter.

        Args:
            table_asset: The TableAsset to which the ColumnSplitter will be applied.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        datasource: SQLDatasource = table_asset.datasource
        engine: sqlalchemy.engine.Engine = datasource.get_engine()
        inspector: sqlalchemy.engine.Inspector = sqlalchemy.inspect(engine)

        columns: list[dict[str, Any]] = inspector.get_columns(
            table_name=table_asset.table_name, schema=table_asset.schema_name
        )
        column_names: list[str] = [column["name"] for column in columns]
        if self.column_name not in column_names:
            raise TestConnectionError(
                f'The column "{self.column_name}" was not found in table "{table_asset.qualified_name}"'
            )


class ColumnSplitterYear(_ColumnSplitter):
    method_name: Literal["split_on_year"] = "split_on_year"

    @property
    def param_names(self) -> List[str]:
        return ["year"]

    def splitter_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name}

    def batch_request_options_to_batch_spec_kwarg_identifiers(
        self, options: BatchRequestOptions
    ) -> Dict[str, Any]:
        return _request_options_to_datetime_batch_spec_kwargs(self, options)


class ColumnSplitterYearAndMonth(_ColumnSplitter):
    method_name: Literal["split_on_year_and_month"] = "split_on_year_and_month"

    @property
    def param_names(self) -> List[str]:
        return ["year", "month"]

    def splitter_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name}

    def batch_request_options_to_batch_spec_kwarg_identifiers(
        self, options: BatchRequestOptions
    ) -> Dict[str, Any]:
        return _request_options_to_datetime_batch_spec_kwargs(self, options)


class ColumnSplitterYearAndMonthAndDay(_ColumnSplitter):
    method_name: Literal[
        "split_on_year_and_month_and_day"
    ] = "split_on_year_and_month_and_day"

    @property
    def param_names(self) -> List[str]:
        return ["year", "month", "day"]

    def splitter_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name}

    def batch_request_options_to_batch_spec_kwarg_identifiers(
        self, options: BatchRequestOptions
    ) -> Dict[str, Any]:
        return _request_options_to_datetime_batch_spec_kwargs(self, options)


class ColumnSplitterDatetimePart(_ColumnSplitter):
    datetime_parts: List[str]
    column_name: str
    method_name: Literal["split_on_date_parts"] = "split_on_date_parts"

    @property
    def param_names(self) -> List[str]:
        return self.datetime_parts

    def splitter_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name, "date_parts": self.param_names}

    def batch_request_options_to_batch_spec_kwarg_identifiers(
        self, options: BatchRequestOptions
    ) -> Dict[str, Any]:
        return _request_options_to_datetime_batch_spec_kwargs(self, options)

    @pydantic.validator("datetime_parts", each_item=True)
    def _check_param_name_allowed(cls, v: str):
        from great_expectations.execution_engine.split_and_sample.data_splitter import (
            DatePart,
        )

        allowed_date_parts = [part.value for part in DatePart]
        assert (
            v in allowed_date_parts
        ), f"Only the following param_names are allowed: {allowed_date_parts}"
        return v


DatetimeColumnSplitter = Union[
    ColumnSplitterYear,
    ColumnSplitterYearAndMonth,
    ColumnSplitterYearAndMonthAndDay,
    ColumnSplitterDatetimePart,
]


def _request_options_to_datetime_batch_spec_kwargs(
    splitter: DatetimeColumnSplitter, options: BatchRequestOptions
) -> Dict[str, Any]:
    identifiers: Dict = {}
    for part in splitter.param_names:
        if part not in options:
            raise ValueError(
                f"'{part}' must be specified in the batch request options to create a batch identifier"
            )
        identifiers[part] = options[part]
    return {splitter.column_name: identifiers}


class ColumnSplitterColumnValue(_ColumnSplitter):
    method_name: Literal["split_on_column_value"] = "split_on_column_value"

    @property
    def param_names(self) -> List[str]:
        return [self.column_name]

    def splitter_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name}

    def batch_request_options_to_batch_spec_kwarg_identifiers(
        self, options: BatchRequestOptions
    ) -> Dict[str, Any]:
        if self.column_name not in options:
            raise ValueError(
                f"'{self.column_name}' must be specified in the batch request options to create a batch identifier"
            )
        return {self.column_name: options[self.column_name]}


class ColumnSplitterDividedInteger(_ColumnSplitter):
    divisor: int
    method_name: Literal["split_on_divided_integer"] = "split_on_divided_integer"

    @property
    def param_names(self) -> List[str]:
        return ["quotient"]

    def splitter_method_kwargs(self) -> Dict[str, Any]:
        return {"column_name": self.column_name, "divisor": self.divisor}

    def batch_request_options_to_batch_spec_kwarg_identifiers(
        self, options: BatchRequestOptions
    ) -> Dict[str, Any]:
        if "quotient" not in options:
            raise ValueError(
                "'quotient' must be specified in the batch request options to create a batch identifier"
            )
        return {self.column_name: options["quotient"]}


ColumnSplitter = Union[
    ColumnSplitterColumnValue,
    ColumnSplitterDividedInteger,
    ColumnSplitterDatetimePart,
    ColumnSplitterYearAndMonthAndDay,
    ColumnSplitterYearAndMonth,
    ColumnSplitterYear,
]


class _SQLAsset(DataAsset):
    # Instance fields
    type: str = pydantic.Field("_sql_asset")
    column_splitter: Optional[ColumnSplitter] = None
    name: str

    # Internal attributes
    _datasource: SQLDatasource = pydantic.PrivateAttr()

    def batch_request_options_template(
        self,
    ) -> BatchRequestOptions:
        """A BatchRequestOptions template for build_batch_request.

        Returns:
            A BatchRequestOptions dictionary with the correct shape that build_batch_request
            will understand. All the option values are defaulted to None.
        """
        template: BatchRequestOptions = {}
        if not self.column_splitter:
            return template
        return {p: None for p in self.column_splitter.param_names}

    def _add_splitter(self, column_splitter: ColumnSplitter):
        self.column_splitter = column_splitter
        self.test_column_splitter_connection()
        return self

    def add_splitter_year(
        self,
        column_name: str,
    ) -> _SQLAsset:
        """Associates a year splitter with this sql data asset.
        Args:
            column_name: A column name of the date column where year and month will be parsed out.
        Returns:
            This sql asset so we can use this method fluently.
        """
        return self._add_splitter(
            ColumnSplitterYear(method_name="split_on_year", column_name=column_name)
        )

    def add_splitter_year_and_month(
        self,
        column_name: str,
    ) -> _SQLAsset:
        """Associates a year, month splitter with this sql asset.
        Args:
            column_name: A column name of the date column where year and month will be parsed out.
        Returns:
            This sql asset so we can use this method fluently.
        """
        return self._add_splitter(
            ColumnSplitterYearAndMonth(
                method_name="split_on_year_and_month", column_name=column_name
            )
        )

    def add_splitter_year_and_month_and_day(
        self,
        column_name: str,
    ) -> _SQLAsset:
        """Associates a year, month, day splitter with this sql asset.
        Args:
            column_name: A column name of the date column where year and month will be parsed out.
        Returns:
            This sql asset so we can use this method fluently.
        """
        return self._add_splitter(
            ColumnSplitterYearAndMonthAndDay(
                method_name="split_on_year_and_month_and_day", column_name=column_name
            )
        )

    def add_splitter_datetime_part(
        self, column_name: str, datetime_parts: List[str]
    ) -> _SQLAsset:
        return self._add_splitter(
            ColumnSplitterDatetimePart(
                method_name="split_on_date_parts",
                column_name=column_name,
                datetime_parts=datetime_parts,
            )
        )

    def add_splitter_column_value(self, column_name: str) -> _SQLAsset:
        return self._add_splitter(
            ColumnSplitterColumnValue(
                method_name="split_on_column_value",
                column_name=column_name,
            )
        )

    def add_splitter_divided_integer(self, column_name: str, divisor: int) -> _SQLAsset:
        return self._add_splitter(
            ColumnSplitterDividedInteger(
                method_name="split_on_divided_integer",
                column_name=column_name,
                divisor=divisor,
            )
        )

    def test_connection(self) -> None:
        pass

    def test_column_splitter_connection(self) -> None:
        pass

    @staticmethod
    def _does_matches_request_options(
        candidate: Dict, requested_options: BatchRequestOptions
    ) -> bool:
        for k, v in requested_options.items():
            if v is not None and candidate[k] != v:
                return False
        return True

    def _fully_specified_batch_requests(
        self, batch_request: BatchRequest
    ) -> List[BatchRequest]:
        """Populates a batch requests unspecified params producing a list of batch requests."""
        if self.column_splitter is None:
            # Currently batch_request.options is complete determined by the presence of a
            # column splitter. If column_splitter is None, then there are no specifiable options
            # so we return early. Since the passed in batch_request is verified, it must be the
            # empty, ie {}.
            # In the future, if there are options that are not determined by the column splitter
            # this check will have to be generalized.
            return [batch_request]

        batch_requests: List[BatchRequest] = []
        # We iterate through all possible batches as determined by the column splitter
        for params in self.column_splitter.param_defaults(self):
            # If the params from the column splitter don't match the batch request options
            # we don't create this batch.
            if not _SQLAsset._does_matches_request_options(
                params, batch_request.options
            ):
                continue
            options = copy.deepcopy(batch_request.options)
            options.update(params)
            batch_requests.append(
                BatchRequest(
                    datasource_name=batch_request.datasource_name,
                    data_asset_name=batch_request.data_asset_name,
                    options=options,
                )
            )
        return batch_requests

    def get_batch_list_from_batch_request(
        self, batch_request: BatchRequest
    ) -> List[Batch]:
        """A list of batches that match the BatchRequest.

        Args:
            batch_request: A batch request for this asset. Usually obtained by calling
                build_batch_request on the asset.

        Returns:
            A list of batches that match the options specified in the batch request.
        """
        self._validate_batch_request(batch_request)

        batch_list: List[Batch] = []
        column_splitter = self.column_splitter
        batch_spec_kwargs: dict[str, str | dict | None]
        for request in self._fully_specified_batch_requests(batch_request):
            batch_metadata = copy.deepcopy(request.options)
            batch_spec_kwargs = self._create_batch_spec_kwargs()
            if column_splitter:
                batch_spec_kwargs["splitter_method"] = column_splitter.method_name
                batch_spec_kwargs[
                    "splitter_kwargs"
                ] = column_splitter.splitter_method_kwargs()
                # mypy infers that batch_spec_kwargs["batch_identifiers"] is a collection, but
                # it is hardcoded to a dict above, so we cast it here.
                cast(Dict, batch_spec_kwargs["batch_identifiers"]).update(
                    column_splitter.batch_request_options_to_batch_spec_kwarg_identifiers(
                        request.options
                    )
                )
            # Creating the batch_spec is our hook into the execution engine.
            batch_spec = SqlAlchemyDatasourceBatchSpec(**batch_spec_kwargs)
            execution_engine: SqlAlchemyExecutionEngine = (
                self.datasource.get_execution_engine()
            )
            data, markers = execution_engine.get_batch_data_and_markers(
                batch_spec=batch_spec
            )

            # batch_definition (along with batch_spec and markers) is only here to satisfy a
            # legacy constraint when computing usage statistics in a validator. We hope to remove
            # it in the future.
            # imports are done inline to prevent a circular dependency with core/batch.py
            from great_expectations.core import IDDict
            from great_expectations.core.batch import BatchDefinition

            batch_definition = BatchDefinition(
                datasource_name=self.datasource.name,
                data_connector_name="experimental",
                data_asset_name=self.name,
                batch_identifiers=IDDict(batch_spec["batch_identifiers"]),
                batch_spec_passthrough=None,
            )

            # Some pydantic annotations are postponed due to circular imports.
            # Batch.update_forward_refs() will set the annotations before we
            # instantiate the Batch class since we can import them in this scope.
            Batch.update_forward_refs()
            batch_list.append(
                Batch(
                    datasource=self.datasource,
                    data_asset=self,
                    batch_request=request,
                    data=data,
                    metadata=batch_metadata,
                    legacy_batch_markers=markers,
                    legacy_batch_spec=batch_spec,
                    legacy_batch_definition=batch_definition,
                )
            )
        self.sort_batches(batch_list)
        return batch_list

    def build_batch_request(
        self, options: Optional[BatchRequestOptions] = None
    ) -> BatchRequest:
        """A batch request that can be used to obtain batches for this DataAsset.

        Args:
            options: A dict that can be used to limit the number of batches returned from the asset.
                The dict structure depends on the asset type. A template of the dict can be obtained by
                calling batch_request_options_template.

        Returns:
            A BatchRequest object that can be used to obtain a batch list from a Datasource by calling the
            get_batch_list_from_batch_request method.
        """
        if options is not None and not self._valid_batch_request_options(options):
            allowed_keys = set(self.batch_request_options_template().keys())
            actual_keys = set(options.keys())
            raise gx_exceptions.InvalidBatchRequestError(
                "Batch request options should only contain keys from the following set:\n"
                f"{allowed_keys}\nbut your specified keys contain\n"
                f"{actual_keys.difference(allowed_keys)}\nwhich is not valid.\n"
            )
        return BatchRequest(
            datasource_name=self.datasource.name,
            data_asset_name=self.name,
            options=options or {},
        )

    def _validate_batch_request(self, batch_request: BatchRequest) -> None:
        """Validates the batch_request has the correct form.

        Args:
            batch_request: A batch request object to be validated.
        """
        if not (
            batch_request.datasource_name == self.datasource.name
            and batch_request.data_asset_name == self.name
            and self._valid_batch_request_options(batch_request.options)
        ):
            expect_batch_request_form = BatchRequest(
                datasource_name=self.datasource.name,
                data_asset_name=self.name,
                options=self.batch_request_options_template(),
            )
            raise gx_exceptions.InvalidBatchRequestError(
                "BatchRequest should have form:\n"
                f"{pf(dataclasses.asdict(expect_batch_request_form))}\n"
                f"but actually has form:\n{pf(dataclasses.asdict(batch_request))}\n"
            )

    def _create_batch_spec_kwargs(self) -> dict[str, Any]:
        """Creates batch_spec_kwargs used to instantiate a SqlAlchemyDatasourceBatchSpec

        This is called by get_batch_list_from_batch_request to generate the batches.

        Returns:
            A dictionary that will be passed to SqlAlchemyDatasourceBatchSpec(**returned_dict)
        """
        raise NotImplementedError

    def as_selectable(self) -> sqlalchemy.sql.Selectable:
        """Returns a Selectable that can be used to query this data

        Returns:
            A Selectable that can be used in a from clause to query this data
        """
        raise NotImplementedError


class QueryAsset(_SQLAsset):
    # Instance fields
    type: Literal["query"] = "query"
    query: str

    @pydantic.validator("query")
    def query_must_start_with_select(cls, v: str):
        query = v.lstrip()
        if not (query.upper().startswith("SELECT") and query[6].isspace()):
            raise ValueError("query must start with 'SELECT' followed by a whitespace.")
        return v

    def as_selectable(self) -> sqlalchemy.sql.Selectable:
        """Returns the Selectable that is used to retrieve the data.

        This can be used in a subselect FROM clause for queries against this data.
        """
        import sqlalchemy as sa

        return sa.select(sa.text(self.query.lstrip()[6:])).subquery()

    def _create_batch_spec_kwargs(self) -> dict[str, Any]:
        return {
            "data_asset_name": self.name,
            "query": self.query,
            "temp_table_schema_name": None,
            "batch_identifiers": {},
        }


class TableAsset(_SQLAsset):
    # Instance fields
    type: Literal["table"] = "table"
    table_name: str
    schema_name: Optional[str] = None

    @property
    def qualified_name(self) -> str:
        return (
            f"{self.schema_name}.{self.table_name}"
            if self.schema_name
            else self.table_name
        )

    def test_connection(self) -> None:
        """Test the connection for the TableAsset.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        datasource: SQLDatasource = self.datasource
        engine: sqlalchemy.engine.Engine = datasource.get_engine()
        inspector: sqlalchemy.engine.Inspector = sqlalchemy.inspect(engine)

        if self.schema_name and self.schema_name not in inspector.get_schema_names():
            raise TestConnectionError(
                f'Attempt to connect to table: "{self.qualified_name}" failed because the schema '
                f'"{self.schema_name}" does not exist.'
            )

        table_exists = sqlalchemy.inspect(engine).has_table(
            table_name=self.table_name,
            schema=self.schema_name,
        )
        if not table_exists:
            raise TestConnectionError(
                f'Attempt to connect to table: "{self.qualified_name}" failed because the table '
                f'"{self.table_name}" does not exist.'
            )

    def test_column_splitter_connection(self) -> None:
        if self.column_splitter:
            self.column_splitter.test_connection(table_asset=self)

    def as_selectable(self) -> sqlalchemy.sql.Selectable:
        """Returns the table as a sqlalchemy Selectable.

        This can be used in a from clause for a query against this data.
        """
        import sqlalchemy as sa

        return sa.text(self.table_name)

    def _create_batch_spec_kwargs(self) -> dict[str, Any]:
        return {
            "type": "table",
            "data_asset_name": self.name,
            "table_name": self.table_name,
            "schema_name": self.schema_name,
            "batch_identifiers": {},
        }


class SQLDatasource(Datasource):
    """Adds a generic SQL datasource to the data context.

    Args:
        name: The name of this datasource.
        connection_string: The SQLAlchemy connection string used to connect to the database.
            For example: "postgresql+psycopg2://postgres:@localhost/test_database"
        assets: An optional dictionary whose keys are SQL DataAsset names and whose values
            are SQL DataAsset objects.
    """

    # class var definitions
    asset_types: ClassVar[List[Type[DataAsset]]] = [TableAsset, QueryAsset]

    # right side of the operator determines the type name
    # left side enforces the names on instance creation
    type: Literal["sql"] = "sql"
    connection_string: str
    # We need to explicitly add each asset type to the Union due to how
    # deserialization is implemented in our pydantic base model.
    assets: Dict[str, Union[TableAsset, QueryAsset]] = {}

    # private attrs
    _cached_connection_string: str = pydantic.PrivateAttr("")
    _engine: Union[sqlalchemy.engine.Engine, None] = pydantic.PrivateAttr(None)

    @property
    def execution_engine_type(self) -> Type[SqlAlchemyExecutionEngine]:
        """Returns the default execution engine type."""
        from great_expectations.execution_engine import SqlAlchemyExecutionEngine

        return SqlAlchemyExecutionEngine

    def get_engine(self) -> sqlalchemy.engine.Engine:
        if self.connection_string != self._cached_connection_string or not self._engine:
            # validate that SQL Alchemy was successfully imported and attempt to create an engine
            if SQLALCHEMY_IMPORTED:
                try:
                    self._engine = sqlalchemy.create_engine(self.connection_string)
                except Exception as e:
                    # connection_string has passed pydantic validation, but still fails to create a sqlalchemy engine
                    # one possible case is a missing plugin (e.g. psycopg2)
                    raise SQLDatasourceError(
                        "Unable to create a SQLAlchemy engine from "
                        f"connection_string: {self.connection_string} due to the "
                        f"following exception: {str(e)}"
                    ) from e
                self._cached_connection_string = self.connection_string
            else:
                raise SQLDatasourceError(
                    "Unable to create SQLDatasource due to missing sqlalchemy dependency."
                )
        return self._engine

    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the SQLDatasource.

        Args:
            test_assets: If assets have been passed to the SQLDatasource, whether to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        try:
            engine: sqlalchemy.engine.Engine = self.get_engine()
            engine.connect()
        except Exception as e:
            raise TestConnectionError(
                "Attempt to connect to datasource failed with the following error message: "
                f"{str(e)}"
            ) from e
        if self.assets and test_assets:
            for asset in self.assets.values():
                asset._datasource = self
                asset.test_connection()

    def add_table_asset(
        self,
        name: str,
        table_name: str,
        schema_name: Optional[str] = None,
        order_by: Optional[BatchSortersDefinition] = None,
    ) -> TableAsset:
        """Adds a table asset to this datasource.

        Args:
            name: The name of this table asset.
            table_name: The table where the data resides.
            schema_name: The schema that holds the table.
            order_by: A list of BatchSorters or BatchSorter strings.

        Returns:
            The TableAsset that is added to the datasource.
        """
        order_by_sorters: list[BatchSorter] = self.parse_order_by_sorters(
            order_by=order_by
        )
        asset = TableAsset(
            name=name,
            table_name=table_name,
            schema_name=schema_name,
            order_by=order_by_sorters,
        )
        return self.add_asset(asset)

    def add_query_asset(
        self,
        name: str,
        query: str,
        order_by: Optional[BatchSortersDefinition] = None,
    ) -> QueryAsset:
        """Adds a query asset to this datasource.

        Args:
            name: The name of this table asset.
            query: The SELECT query to selects the data to validate. It must begin with the "SELECT".
            order_by: A list of BatchSorters or BatchSorter strings.

        Returns:
            The QueryAsset that is added to the datasource.
        """
        order_by_sorters: list[BatchSorter] = self.parse_order_by_sorters(
            order_by=order_by
        )
        asset = QueryAsset(
            name=name,
            query=query,
            order_by=order_by_sorters,
        )
        return self.add_asset(asset)
